"""Life insurance - Computes present values of insurance benefits

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
from typing import Callable, Any, Tuple, List
import math
import matplotlib.pyplot as plt
import numpy as np
from actuarialmath.fractional import Fractional

class Insurance(Fractional):
    """Compute expected present values of life insurance"""

    def E_x(self, x: int, s: int = 0, t: int = 1, endowment: int = 1,
            moment: int = 1) -> float:
        """Pure endowment: t_E_x

        Args:
          x : age of selection
          s : years after selection
          t : term of pure endowment
          endowment : amount of pure endowment
          moment : compute first or second moment
        """
        if t < 0:  # t infinite => EPV(t) = 0
            return 0   
        if t == 0:    # t = 0 => EPV(0) = 1
            return 1
        t = self.max_term(x+s, t)
        t_p_x = self.p_x(x, s=s, t=t)
        if moment == self.VARIANCE:  # Bernoulli shortcut for variance
            return self.interest.v_t(t)**2 * t_p_x * (1 - t_p_x) * endowment**2
        return (endowment * self.interest.v_t(t))**moment * t_p_x

    def A_x(self, x: int, s: int = 0, t: int = Fractional.WHOLE, u: int = 0,
            benefit: Callable = lambda x,t: 1., endowment: float = 0.,
            moment: int = 1, discrete: bool = True) -> float:
        """Numerically compute EPV of insurance from basic survival functions

        Args:
          x : age of selection
          s : years after selection
          u : year deferred
          t : term of insurance
          benefit : benefit as a function of age and year
          endowment : amount of endowment for endowment insurance
          moment : compute first or second moment
          discrete : benefit paid yearend (True) or moment of death (False)
        """
        assert moment >= 1
        if t >=0 and endowment > 0:
            E = self.E_x(x, s=s, t=t+u, moment=moment) * endowment
        else:
            E = 0
        t = self.max_term(x+s+u, t=t)
        if discrete:
            A = sum([(benefit(x+s, k+1)*self.interest.v_t(k+1))**moment 
                     * self.q_x(x, s=s, u=k) for k in range(u, t+u)])
        else:   # use continous first principles
            Z = lambda t: ((benefit(x+s, t+u) * self.interest.v_t(t+u))**moment 
                           * self.f(x, s, t+u))
            A = self.integral(Z, 0, t)
        return A + E

    @staticmethod
    def insurance_variance(A2: float, A1: float, b: float = 1) -> float:
        """Compute variance of insurance given moments and benefit

        Args:
          A2 : second moment of insurance r.v.
          A1 : first moment of insurance r.v.
          b : benefit amount
        """
        return b**2 * (A2 - A1**2)

    def whole_life_insurance(self, x: int, s: int = 0, moment: int = 1, 
                             b: int = 1, discrete: bool = True) -> float:
        """Whole life insurance: A_x

        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if moment == self.VARIANCE:
            A2 = self.whole_life_insurance(x, s=s, moment=2, discrete=discrete)
            A1 = self.whole_life_insurance(x, s=s, discrete=discrete)**2
            return self.insurance_variance(A2=A2, A1=A1, b=b)
        return self.A_x(x, s=s, t=self.WHOLE, benefit=lambda x,t: b, 
                        moment=moment, discrete=discrete)

    def term_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                       moment: int = 1, discrete: bool = True) -> float:
        """Term life insurance: A_x:t^1

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if moment == self.VARIANCE:
            A2 = self.term_insurance(x, s=s, t=t, moment=2, discrete=discrete)
            A2 = self.term_insurance(x, s=s, t=t, discrete=discrete)**2
            return self.insurance_variance(A2=A2, A1=A1, b=b)
        A = self.whole_life_insurance(x, s=s, b=b, moment=moment, 
                                      discrete=discrete)
        if t < 0 or self.max_term(x+s, t=t) < t:
            return A
        E = self.E_x(x, s=s, t=t, moment=moment)
        A -= E * self.whole_life_insurance(x, s=s+t, b=b, moment=moment,
                                            discrete=discrete)
        return A

    def deferred_insurance(self, x: int, s: int = 0, u: int = 0, 
                           t: int = Fractional.WHOLE, b: int = 1, 
                           moment: int = 1, discrete: bool = True) -> float:
        """Deferred insurance n|_A_x:t^1 = discounted term or whole life

        Args:
          x : age of selection
          s : years after selection
          u : year deferred
          t : term of insurance
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if self.max_term(x+s, u) < u:
            return 0.        
        if moment == self.VARIANCE:
            A2 = self.deferred_insurance(x, s=s, t=t, u=u, moment=2, 
                                         discrete=discrete)
            A1 = self.deferred_insurance(x, s=s, t=t, u=u, discrete=discrete)
            return self.insurance_variance(A2=A2, A1=A1, b=b)
        E = self.E_x(x, s=s, t=u, moment=moment)
        A = self.term_insurance(x, s=s+u, t=t, b=b, moment=moment, 
                                discrete=discrete)
        return E * A  # discount insurance by moment*force of interest

    def endowment_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                            endowment: int = -1, moment: int = 1, 
                            discrete: bool = True) -> float:
        """Endowment insurance: A_x^1:t = term insurance + pure endowment

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit
          endowment : amount of endowment paid at end of term if survive
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if moment == self.VARIANCE:
            A2 = self.endowment_insurance(x, s=s, t=t, endowment=endowment, 
                                          b=b, moment=2, discrete=discrete)
            A1 = self.endowment_insurance(x, s=s, t=t, endowment=endowment,
                                          b=b, discrete=discrete)
            return self.insurance_variance(A2=A2, A1=A1, b=b)
        E = self.E_x(x, s=s, t=t, moment=moment)
        A = self.term_insurance(x, s=s, t=t, b=b, moment=moment,
                                discrete=discrete)
        return A + E * (b if endowment < 0 else endowment)**moment

    def increasing_insurance(self, x: int, s: int = 0, t: int =
                             Fractional.WHOLE, 
                             b: int = 1, discrete: bool = True) -> float:
        """Increasing life insurance: (IA)_x

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit in first year
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        return self.A_x(x, s=s, t=t, benefit=lambda x,t: t * b, 
                        discrete=discrete)

    def decreasing_insurance(self, x, s: int = 0, t: int = 1, b: int = 1,
                             discrete: bool = True) -> float:
        """Decreasing life insurance: (DA)_x

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit in first year
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        assert t > 0  # decreasing must be term insurance
        A = self.term_insurance(x, t=t, b=b, discrete=discrete)
        n = t + int(discrete)   #  (DA)_x:n + (IA)_x:n = (n+1) A^1_x:n
        return A*n - self.increasing_insurance(x, s=s, t=t, b=b, 
                                               discrete=discrete)

    #
    # Insurance random variable: Z(t)
    #
    def Z_t(self, x: int, prob: float, discrete: bool = True) -> float:
        """T_x given percentile of the PV of WL or Term insurance, i.e. r.v. Z(t)

        Args:
          x : age initially insured
          prob : desired probability threshold
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        assert prob < 1.0
        t = Insurance.solve(lambda t: self.S(x, 0, t), target=prob, grid=50)
        return math.floor(t) if discrete else t    # opposite of annuity

    def Z_from_t(self, t: float, discrete: bool = True) -> float:
        """PV of insurance payment Z(t), given T_x (or K_x if discrete)

        Args:
          t : year of death
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        return self.interest.v_t((math.floor(t) + 1) if discrete else t)

    def Z_to_t(self, Z: float) -> float:
        """T_x s.t. PV of insurance payment is Z

        Args:
          Z : Present value of benefit paid
        """
        #t = Insurance.solve(lambda t: self.Z_from_t(t), Z, self._MAXAGE/2)
        t = math.log(Z) / math.log(self.interest.v)
        return t

    def Z_from_prob(self, x: int, prob: float, discrete: bool = True) -> float:
        """Percentile of insurance PV r.v. Z, given probability

        Args:
          x : age initially insured
          prob : threshold for probability of survival
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        t = self.Z_t(45, prob, discrete=discrete)   # opposite of annuity!
        return self.Z_from_t(t, discrete=discrete)  # z is WL or Term Insurance

    def Z_to_prob(self, x: int, Z: float) -> float:
        """Cumulative density of insurance PV r.v. Z, given percentile value

        Args:
          x : age initially insured
          Z : present value of benefit paid
        """
        t = self.Z_to_t(Z) 
        return self.S(x, 0, t)      # z is WL or Term Insurance

    def Z_x(self, x, s: int = 0, t: int = 1, discrete: bool = True):
        """EPV of year t insurance death benefit for life aged [x]+s: b_x[s]+s(t)

        Args:
          x : age of selection
          s : years after selection
          t : year of benefit
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        assert t > 0
        if discrete:
            u = math.ceil(t) - 1
            return self.q_x(x, s=s, u=u) * self.interest.v_t(u+1)
        else:
            return self.f_r(x, s=s, t=t) * self.interest.v_t(t)

    def Z_plot(self, x: int, s: int = 0, stop : int = 0,
               benefit: Callable = lambda x,k: 1,
               T: float | None = None,
               discrete: bool = True,
               ax: Any = None,
               title: str | None = None,
               color: str ='r')-> float | None:
        """Plot of PV of insurance r.v. Z vs t

        Args:
          x : age of selection
          s : years after selection
          stop : time to end plot
          benefit : benefit as a function of selection age and time
          discrete  discrete or continuous insurance
          ax : figure object to plot in
          color : color to plot
          title : title of plot 
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        K = 'K' if discrete else 'T'
        stop = stop or self._MAXAGE - (x + s)
        step = 1 if discrete else stop / 1000.
        steps = np.arange(0, stop + step, step)
        
        # plot PV benefit values
        z = [benefit(x, s+t) * self.Z_from_t(t, discrete=discrete) for t in steps]
        if T is None:
            ax.bar(steps, z, width=step, alpha=0.5, color=color)
            Z = None
        else:
            # plot Z(t)
            ax.step(steps, z, ':', c=color, where='pre' if discrete else 'post')
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            yjig = (ymax - ymin) / 50
            xjig = (xmax - xmin) / 50
            
            # indicate PV of benefit Z(T*)
            Z = self.Z_from_t(T, discrete=discrete) * benefit(self._MINAGE, T)
            ax.plot(T, Z, c=color, marker='o')
            ax.text(T, Z + yjig, f"Z*={Z:.2f}", c=color)

            # indicate given time of death T*
            ax.vlines(T, ymin, Z, colors='g', linestyles=':')
            ax.text(T + xjig, ymin, f"${K}_x$={T:.2f}", c='g')

            # indicate corresponding S(T*)
            p = self.S(x, 0, T)     # S(t): note that is opposite of annuity
            ax.hlines(Z, T, xmax, colors='g', linestyles=':')
            ax.text(xmax, Z - yjig, f"Prob={p:.3f}", c='g', va='top', ha='right')
        ax.set_title(f"PV insurance r.v. $Z({K}_{{{x if x else 'x'}}})$"
                     if title is None else title)
        ax.set_ylabel(f"$Z({K}_x)$", color=color)
        ax.set_xlabel(f"${K}_x$")
        plt.tight_layout()
        return Z

    def Z_curve(self, x: int, s: int = 0, stop: int = 0,
                benefit: Callable = lambda x, k: 1,
                T: float | None = None,
                discrete: bool = True,
                title: str | None = None,
                ax: Any = None):
        """Plot PV of insurance r.v. Z(t) and survival probability vs time t

        Args:
          x : age of selection
          s : years after selection
          stop : end at time t, inclusive
          benefit : benefit as a function of selection age and time
          discrete : discrete or continuous insurance
          T : point in time to indicate benefit and survival probability
          ax : figure object to plot in
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)        
        K = 'K' if discrete else 'T'
        stop = stop or self._MAXAGE - (x + s)
        step = 1 # if discrete else stop/1000.
        steps = range(stop + step)
        p = [self.p_x(x=x, s=s, t=t) for t in steps]

        # plot survival probabilities in secondary axis
        bx = ax.twinx()
        bx.bar(steps, p, color='g', alpha=.2, width=step, align='edge')
        bx.set_ylabel(f"$S({K})$", color='g')
        bx.tick_params(axis='y', colors='g')

        # plot benefit values in primary axis
        z = [benefit(x, s+t) * self.Z_from_t(t, discrete=discrete) for t in steps]
        ax.step(steps, z, ':', c='r', where='pre' if discrete else 'post')
        ax.set_ylabel(f"Z({K})", color='r')
        ax.tick_params(axis='y', colors='r')

        if T is not None:   # plot PV benefit value and survival prob at given T
            Z = self.Z_from_t(T, discrete=discrete) * benefit(self._MINAGE, T)
            label1, = ax.plot(T, Z, c='r', marker='o',
                              label=f"Z({K}*={T:.2f}): {Z:.2f}")
            ax.legend(handles=[label1], loc='center left')
            
            prob = self.S(x, 0, T)      # note: is opposite of annuity
            label2, = bx.plot(T, prob, c='g', marker='o',
                               label=f"Pr[{K}>{K}*]<={prob:.3f}")
            bx.legend(handles=[label2], loc='center right')
        ax.set_title(title if title is not None else
                     f"PV benefit $Z({K})$ and survival probability $S({K})$")
        ax.set_xlabel(f"${K}$")
        plt.tight_layout()

    
if __name__ == "__main__":

    from actuarialmath.sult import SULT
    life = SULT()
    life.Z_curve(x=20, stop=80, T=life.Z_t(x=20, prob=0.5),
                 title="PV insurance benefit if (x) survives median lifetime")
    
    print("SOA Question 4.18  (A) 81873 ")
    def f(x,s,t): return 0.1 if t < 2 else 0.4*t**(-2)
    life = Insurance().set_interest(delta=0.05)\
                      .set_survival(f=f, maxage=10)
    def benefit(x,t): return 0 if t < 2 else 100000
    prob = 0.9 - life.q_x(x=0, t=2)
    T = life.Z_t(x=0, prob=prob)
    life.Z_curve(x=0, T=T, benefit=benefit, discrete=False)
    Z = life.Z_from_t(T) * benefit(0, T)
    print(Z)
    
    # Example: plot Z vs T
    life = Insurance().set_interest(delta=0.06)\
                      .set_survival(mu=lambda *x: 0.04)
    prob = 0.8
    x = 20
    discrete = True
    t = life.Z_t(x, prob, discrete=discrete)
    Z = life.Z_from_prob(x, prob=prob, discrete=discrete)
    print(t, life.Z_to_t(Z))
    print(Z, life.Z_from_t(t, discrete=discrete))
    print(prob, life.Z_to_prob(x, Z=Z))
    life.Z_plot(x, T=t, discrete=discrete)
    plt.show()
    

    print("SOA Question 4.10:  (D)")
    x = 0
    life = Insurance().set_interest(i=0.0)\
                      .set_survival(S=lambda x,s,t: 1, maxage=x+40)
    def expected(x, t):  # true E[Z]
        if 10 <= t <= 20: return life.interest.v_t(t)
        elif 20 < t <= 30: return 2 * life.interest.v_t(t)
        else: return 0
        
    def A(x, t):  # Z_x+k (t-k)
        return life.interest.v_t(t - x) * (t > x)
    
    benefits=[lambda x,t: (life.E_x(x, t=10) * A(x+10, t)
                             + life.E_x(x, t=20)* A(x+20, t)
                             - life.E_x(x, t=30) * A(x+30, t)),
              lambda x,t: (A(x, t)
                             + life.E_x(x, t=20) * A(x+20, t)
                             - 2 * life.E_x(x, t=30) * A(x+30, t)),
              lambda x,t: (life.E_x(x, t=10) * A(x, t)
                             + life.E_x(x, t=20) * A(x+20, t)
                             - 2 * life.E_x(x, t=30) * A(x+30, t)),
              lambda x,t: (life.E_x(x, t=10) * A(x+10, t)
                             + life.E_x(x, t=20) * A(x+20, t)
                             - 2 * life.E_x(x, t=30) * A(x+30, t)),
              lambda x,t: (life.E_x(x, t=10)
                             * (A(x+10, t)
                                + life.E_x(x+10, t=10) * A(x+20, t)
                                - life.E_x(x+20, t=10) * A(x+30, t)))]
    z = [sum(abs(b(x, t) - expected(x, t)) for t in range(40)) for b in benefits]
    print("ABCDE"[np.argmin(z)])
    
    fig, ax = plt.subplots(3, 2)
    ax = ax.ravel()
    for i, b in enumerate([expected] + benefits):
        life.Z_plot(x, benefit=b, ax=ax[i], color=f"C{i+1}", title=' ')
        ax[i].legend(["(" + "abcde"[i-1] + ")" if i else "Z"])        

    print("SOA Question 6.33:  (B) 0.13")
    life = Insurance().set_survival(mu=lambda x,t: 0.02*t).set_interest(i=0.03)
    x = 0
    var = life.E_x(x, t=15, moment=life.VARIANCE, endowment=10000)
    p = 1- life.portfolio_cdf(mean=0, variance=var, value=50000, N=500)
    print(p)


    print("SOA Question 4.12:  (C) 167")
    cov = Insurance.covariance(a=1.65, b=10.75, ab=0) # Z1 and Z2 nonoverlapping
    var = Insurance.variance(a=2, b=1, var_a=46.75, var_b=50.78, cov_ab=cov)
    print(var)

    print("SOA Question 4.11:  (A) 143385")
    A1 = 528/1000   # E[Z1]  term insurance
    C1 = 0.209      # E[pure_endowment]
    C2 = 0.136      # E[pure_endowment^2]
    def fun(A2):
        B1 = A1 + C1   # endowment = term + pure_endowment
        B2 = A2 + C2   # double force of interest
        return Insurance.insurance_variance(A2=B2, A1=B1)
    A2 = Insurance.solve(fun, target=15000/(1000*1000), grid=[143400, 279300])
    var = Insurance.insurance_variance(A2=A2, A1=A1, b=1000)
    print(var)

    print("SOA Question 4.15  (E) 0.0833 ")
    life = Insurance().set_survival(mu=lambda *x: 0.04)\
                      .set_interest(delta=0.06)
    benefit = lambda x,t: math.exp(0.02*t)
    A1 = life.A_x(0, benefit=benefit, discrete=False)
    A2 = life.A_x(0, moment=2, benefit=benefit, discrete=False)
    var = A2 - A1**2
    print(var)

    print("SOA Question 4.4  (A) 0.036")
    x = 40
    life = Insurance().set_survival(f=lambda *x: 0.025, maxage=x+40)\
                      .set_interest(v_t=lambda t: (1 + .2*t)**(-2))
    benefit = lambda x,t: 1 + .2 * t
    A1 = life.A_x(x, benefit=benefit, discrete=False)
    A2 = life.A_x(x, moment=2, benefit=benefit, discrete=False)
    var = A2 - A1**2
    print(var)

    print("Other examples of usage")
    life = Insurance().set_interest(delta=0.06)\
                      .set_survival(mu=lambda *x: 0.04)
    benefit = lambda x,t: math.exp(0.02 * t)
    A1 = life.A_x(0, benefit=benefit)
    A2 = life.A_x(0, moment=2, benefit=benefit)
    var = A2 - A1**2 
    print(var)  # 0.0833

    life = Insurance().set_interest(delta=0.05)\
                      .set_survival(mu=lambda x,s: 0.03)
    benefit = lambda x,t: math.exp(0.04 * t)
    A = life.A_x(0, benefit=benefit)
    print(A)   # 0.75
    A2 = life.A_x(0, moment=2, benefit=benefit)
    print(A2)  #0.60



"""Mortality Laws: Applies shortcut formulas for Beta, Uniform, Makeham and Gompertz

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
from actuarialmath.reserves import Reserves

class MortalityLaws(Reserves):
    """Apply shortcut formulas for special mortality laws"""

    #
    # Define fractional age actuarial survival functions
    #
    def l_r(self, x: int, s: int = 0, r: float = 0.) -> float:
        """Fractional lives given special mortality law: l_[x]+s+r

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
        """
        return self.l(x, s+r)

    def p_r(self, x: int, s: int = 0, r: float = 0., t: float = 1.) -> float:
        """Fractional age survival probability given special mortality law

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          t : death within next t fractional years
        """
        return self.S(x, s+r, t)

    def q_r(self, x: int, s: int = 0, r: float = 0., t: float = 1., 
            u: float = 0.) -> float:
        """Fractional age deferred mortality given special mortality law

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          u : survive u fractional years, then...
          t : death within next t fractional years
        """
        return self.p_r(x, s=s, r=r, t=u) - self.p_r(x, s=s, r=r, t=t+u)
        
    def mu_r(self, x: int, s: int = 0, r: float = 0.) -> float:
        """Fractional age force of mortality given special mortality law

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
        """
        return self.mu(x, s+r)

    def f_r(self, x: int, s: int = 0, r: float = 0., t: float = 0.0) -> float:
        """fractional age lifetime density given special mortality law

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          t : mortality at fractional year t
        """
        return self.f(x, s+r, t)

    def e_r(self, x: int, s: int = 0, r: float = 0.,
            t: float = Reserves._WHOLE) -> float:
        """Fractional age future lifetime given special mortality law

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          t : limited at t years
        """
        if t < 0:
            t = self._MAXAGE - (x + s + r)
        return self.integrate(lambda t: self.S(x, s+r, t), 0., float(t))

class Beta(MortalityLaws):
    """Shortcuts with beta distribution of deaths (is Uniform when alpha = 1)

    Args:
      omega : maximum age
      alpha : alpha paramter of beta distribution
      lives : assumed starting number of lives for survival function
    """
    
    def __init__(self, omega: int, alpha: float,
                 lives: int = MortalityLaws._RADIX, **kwargs): 
        """Two parameters: alpha and omega, with mu(x) = alpha/(omega-x)"""

        super().__init__(**kwargs)

        def _mu(x: int, s: float) -> float: 
            return alpha / (omega - (x+s))

        def _l(x: int, s: float) -> float:
            return lives * (omega - (x+s))**alpha

        def _S(x: int, s,t : float) -> float:
            return ((omega-(x+s+t))/(omega-(x+s)))**alpha

        def _f(x: int, s,t : float) -> float:
            return alpha / (omega - (x+s))

        self.set_survival(mu=_mu, l=_l, S=_S, f=_f, minage=0, maxage=omega)
        self.omega_ = omega  # store omega parameter
        self.alpha_ = alpha  # store alpha parameter

    def e_r(self, x: int, s: int = 0, t: float = Reserves._WHOLE) -> float:
        """Expectation of future lifetime through fractional age: e_[x]+s:t

        Args:
          x : age of selection
          s : years after selection
          t : limited at t years
        """
        e = (self.omega_ - (x+s)) / (self.alpha_ + 1)
        if t > 0 and self.max_term(x+s, t) > t: # temporary expectation
            return e - self.p_r(x, s=s, t=t) * self.e_r(x, s=s+t)
        return e   # complete expectation

    def e_x(self, x: int, s: int = 0, n: int = MortalityLaws._WHOLE, 
          curtate: bool = False, moment: int = 1) -> float:
        """Shortcut formula for complete expectation

        Args:
          x : age of selection
          s : years after selection
          n : death within next n fractional years
          t : limited at t years
          curtate : whether curtate (True) or complete (False) lifetime
          moment : whether to compute first (1) or second (2) moment
        """
        if n == 0:
            return 0
        if not curtate:
            if moment == 1:
                return self.e_r(x, s=s, t=n)
            if moment == self._VARIANCE and n < 0: # shortcut for complete variance
                return ((self.omega_ - (x + s))
                        / ((self.alpha_ + 1)**2 * (self.alpha_ + 1)))
        return super().__init__(x=x, s=s, n=n, curtate=curtate, moment=moment)

class Uniform(Beta):
    """Shortcuts with uniform distribution of deaths aka DeMoivre's Law

    Args:
      omega : maximum age
      udd : assume UDD (True, default) or CFM (False) between integer ages
    """

    def __init__(self, omega: int, udd: bool = True, **kwargs):
        """One parameter: omega = maxage, with mu(x) = 1/(omega - x)"""
        super().__init__(omega=omega, alpha=1, udd=udd, **kwargs)

    def e_x(self, x: int, s: int = 0, t: int = Beta._WHOLE, 
          curtate: bool = False, moment: int = 1) -> float:
        """Shortcut for expected future lifetime

        Args:
          x : age of selection
          s : years after selection
          t : limited at t years
          curtate : whether curtate (True) or complete (False) lifetime
          moment : whether to compute first (1) or second (2) moment
        """
        if moment in [1, self._VARIANCE] and not curtate:
            if t < 0:
                if moment == self._VARIANCE:
                    return (self.omega_ - x)**2 / 12  # complete shortcut
                else:
                    return (self.omega_ - x) / 2
            elif moment == 1:         # temporary expectation shortcut
                # (Pr[die within n years] * n/2) plus (Pr[survive n years] * n)
                t = self.max_term(x+s, t)
                t_p_x = t / (self.omega_ - x)
                return t_p_x * t  + (1 - t_p_x) * (t / 2)
        return super().__init__(x=x, s=s, t=t, curtate=curtate, moment=moment)

    
    def E_x(self, x: int, s: int = 0, t: int = Beta._WHOLE, 
            moment: int = 1) -> float:
        """Shortcut for Pure Endowment

        Args:
          x : age of selection
          s : years after selection
          t : term of pure endowment
          endowment : amount of pure endowment
          moment : compute first or second moment
        """
        assert moment > 0
        if t == 0:
            return 1.
        if t < 0:
            return 0.
        t = self.max_term(x+s, t)
        t_p_x = (self.omega_ - x - t) / (self.omega_ - x)
        if moment == self._VARIANCE:  # Bernoulli shortcut for variance
            return self.interest.v_t(t)**2 * t_p_x * (1 - t_p_x)
        return (self.interest.v_t(t)**moment * (self.omega_ - x - t) 
                / (self.omega_ - x))

    def whole_life_insurance(self, x: int, s: int = 0, moment: int = 1, 
                             b: int = 1, discrete: bool = True) -> float:
        """Shortcut for whole life insurance

        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """

        if not discrete:
            if moment == Beta._VARIANCE:
                return (self.whole_life_insurance(x, s=s, b=b, moment=2, 
                                                  discrete=False) -
                        self.whole_life_insurance(x, s=s, b=b, moment=1, 
                                                    discrete=False)**2) * b**2
            return self.term_insurance(x, s=s, t=self.omega_-(x+s), b=b,
                                       moment=moment, discrete=False)
        return super().whole_life_insurance(x, s=s, moment=moment, b=b,
                                            discrete=discrete)

    def term_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                       moment: int = 1, discrete: bool = False) -> float:
        """Shortcut for term insurance

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if not discrete and moment in [1, Beta._VARIANCE]:
            if moment == Beta._VARIANCE:
                return (self.term_insurance(x, s=s, b=b, t=t, moment=2, 
                                            discrete=False) -
                        self.term_insurance(x, s=s, b=b, t=t, 
                                            discrete=False)**2) * b**2
            t = self.max_term(x+s, t)                                        
            return (b * (1 - self.interest.v_t(t)) 
                    / (self.interest.delta * (self.omega_ - (x+s))))
        return super().term_insurance(x, s=s, moment=moment, b=b, 
                                      discrete=discrete)

class Makeham(MortalityLaws):
    """Includes element in force of mortality that does not depend on age

    Args:
      A, B, c : parameters of Makeham distribution
    """
    
    def __init__(self, A: float, B: float, c: float, **kwargs):
        assert c > 1, "Makeham requires c > 1"
        assert B > 0, "Makeham requires B > 0"
        assert A >= -B, "Makeham requires A >= -B"        
        super().__init__(**kwargs)
        self.A_ = A
        self.B_ = B
        self.c_ = c

        def _mu(x, s): 
            return A + B * c**(x+s)
        
        def _S(x, s, t):
            return math.exp(-A*t - B*c**(x+s) * (c**t - 1)/math.log(c))

        self.set_survival(mu=_mu, S=_S)

class Gompertz(Makeham):
    """Is Makeham's Law with A = 0

    Args:
      B, c : parameters of Gompertz distribution
    """

    def __init__(self, B: float, c: float):
        """Gompertz's Law is Makeham's Law with A = 0"""
        super().__init__(A=0., B=B, c=c)

if __name__ == "__main__":
    print('Beta')
    life = Beta(omega=100, alpha=0.5)
    print(life.q_x(25, t=1, u=10))     # 0.0072
    print(life.e_x(25))                # 50
    print(Beta(omega=60, alpha=1/3).mu_x(35) * 1000)
    print()

    print('Uniform')
    uniform = Uniform(80).set_interest(delta=0.04)
    print(uniform.whole_life_annuity(20, discrete=False))        # 15.53
    print(uniform.temporary_annuity(20, t=5, discrete=False))   # 4.35
    print(Uniform(161).p_x(70, t=1)) # 0.98901
    print(Uniform(95).e_x(30, t=40, curtate=False)) # 27.692
    print()

    uniform = Uniform(omega=80).set_interest(delta=0.04)
    print(uniform.E_x(20, t=5))  # .7505
    print(uniform.whole_life_insurance(20, discrete=False))  # .3789
    print(uniform.term_insurance(20, t=5, discrete=False))  # .0755
    print(uniform.endowment_insurance(20, t=5, discrete=False))  # .8260
    print(uniform.deferred_insurance(20, u=5, discrete=False))  # .3033
    print()

    print('Gompertz/Makeham')
    life = Gompertz(B=0.000005, c=1.10)
    p = life.p_x(80, t=10)  # 869.4
    print(life.portfolio_percentile(N=1000, mean=p, variance=p*(1-p), prob=0.99)) 

    print(Gompertz(B=0.00027, c=1.1).f_x(50, t=10)) # 0.04839
    life = Makeham(A=0.00022, B=2.7e-6, c=1.124)
    print(life.mu_x(60) * 0.9803)  # 0.00316



"""Woolhouse - 1/mthly insurance and annuities with Woolhouse's approximation

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
from typing import Callable
from actuarialmath.mthly import Mthly
from actuarialmath.annuity import Annuity

class Woolhouse(Mthly):
    """1/m'thly shortcuts with Woolhouse approximation

    Args:
      m : number of payments per year
      life : original fractional survival and mortality functions
      three_term : whether to include (True) or ignore (False) third term
      approximate_mu : function to approximate mu_x for third term
    """
    
    def __init__(self, m: int, life: Annuity, three_term: bool = False,
                 approximate_mu: Callable[[int, int], float] | bool = True):
        super().__init__(m=m, life=life)
        self.three_term = three_term          # whether to include third term
        self.approximate_mu = approximate_mu  # whether to approximate mu

    def mu_x(self, x: int, s: int = 0) -> float:
        """Approximates or computes mu_x for third term

        Args:
          x : age of selection
          s : years after selection
        """
        if self.approximate_mu is True:       # approximate mu
            return -.5 * sum([math.log(self.life.p_x(x, s=s+t)) for t in [0,-1]])
        elif self.approximate_mu is False:
            return self.life.mu_x(x, s=s)     # use exact mu
        else:                                 # apply custom function for mu
            return self.approximate_mu(x, s)

    def whole_life_insurance(self, x: int, s: int = 0, b: int = 1,
                             mu: float | None = 0.) -> float:
        """1/m'thly Woolhouse whole life insurance: A_x
 
        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          mu : value of mu at age x+s
        """
        return b * self.insurance_twin(self.whole_life_annuity(x, s=s, mu=mu))

    def term_insurance(self, x: int, s: int = 0, t: int = Annuity.WHOLE,
                       b: int = 1, mu: float | None = 0., 
                       mu1: float | None = None) -> float:
        """1/m'thly Woolhouse term insurance: A_x:t

        Args:
          x : year of selection
          s : years after selection
          t : term of insurance in years
          b : amount of benefit
          mu : value of mu at age x+s
          mu1 : value of mu at age x+s+t
        """
        A = self.whole_life_insurance(x, s=s, b=b, mu=mu)
        if t < 0 or self.life.max_term(x+s, t) < t:
            return A
        E = self.E_x(x, s=s, t=t)
        return A - E * self.whole_life_insurance(x, s=s+t, b=b, mu=mu1)

    def endowment_insurance(self, x: int, s: int = 0, t: int = Annuity.WHOLE,
                            b: int = 1, endowment: int = -1, 
                            mu: float | None = None) -> float:
        """1/m'thly Woolhouse term insurance: A_x:t

        Args:
          x : year of selection
          s : years after selection
          t : term of insurance in years
          b : amount of benefit
          endowment : amount of endowment
          mu : value of mu at age x+s+t
        """
        if endowment < 0:
            endowment = b
        E = self.E_x(x, s=s, t=t)
        A = self.term_insurance(x, s=s, t=t, b=b, mu=mu)
        return A + E * (b if endowment < 0 else endowment)

    def deferred_insurance(self, x: int, s: int = 0, u: int = 0,
                           t: int = Annuity.WHOLE, b: int = 1,
                           mu: float | None = None, 
                           mu1: float | None = None) -> float:
        """1/m'thly Woolhouse deferred insurance as discounted term or WL

        Args:
          x : year of selection
          s : years after selection
          u : number of years deferred
          t : term of insurance in years
          b : amount of benefit
          mu : value of mu at age x+s+u
          mu1 : value of mu at age x+s+u+t
        """
        if self.life.max_term(x+s, u) < u:
            return 0.
        E = self.E_x(x, s=s, t=u)
        return E * self.term_insurance(x, s=s+u, t=t, b=b, mu=mu, mu1=mu1)

    def whole_life_annuity(self, x: int, s: int = 0, b: int = 1,
                           mu: float | None = None) -> float:
        """1/m'thly Woolhouse whole life annuity: a_x

        Args:
          x : year of selection
          s : years after selection
          b : amount of benefit
          mu : value of mu at age x+s
        """
        a = (self.life.whole_life_annuity(x, s=s, discrete=True) -
             (self.m - 1)/(2 * self.m))
        if self.three_term:
            mu = mu or self.mu_x(x, s)
            a -= (mu + self.life.interest.delta)*(self.m**2 - 1)/(12 * self.m**2)
        return a * b

    def temporary_annuity(self, x: int, s: int = 0, t: int = Annuity.WHOLE, 
                          b: int = 1, mu: float | None = None,
                          mu1: float | None = None) -> float:
        """1/m'thly Woolhouse temporary life annuity: a_x

        Args:
          x : year of selection
          s : years after selection
          t : term of annuity in years
          b : amount of benefit
          mu : value of mu at age x+s
          mu1 : value of mu at age x+s+t
        """
        return self.deferred_annuity(x, s=s, t=t, b=b, mu=mu, mu1=mu1) # u=0

    def deferred_annuity(self, x: int, s: int = 0, t: int = Annuity.WHOLE, 
                         u: int = 0, b: int = 1, mu: float | None = None, 
                        mu1: float | None = None) -> float:
        """1/m'thly Woolhouse deferred life annuity: a_x

        Args:
          x : year of selection
          s : years after selection
          u : years of deferral
          t : term of annuity in years
          mu : value of mu at age x+s+u
          mu1 : value of mu at age x+s+u+t
        """
        a_x = self.whole_life_annuity(x, s=s+u, mu=mu)
        a_xt = self.whole_life_annuity(x, s=s+t+u, mu=mu1) if t > 0 else 0
        a = self.E_x(x, s=s, t=u) * (a_x - self.E_x(x, s=s+u, t=t) * a_xt)
        return a * b

if __name__ == "__main__":
    from actuarialmath.sult import SULT
    from actuarialmath.recursion import Recursion
    from actuarialmath.udd import UDD
    from actuarialmath.policyvalues import Contract
    
    print("SOA Question 7.7:  (D) 1110")
    x = 0
    life = Recursion().set_interest(i=0.05).set_A(0.4, x=x+10)
    a = Woolhouse(m=12, life=life).whole_life_annuity(x+10)
    print(a)
    contract = Contract(premium=0, benefit=10000, renewal_policy=100)
    V = life.gross_future_loss(A=0.4, contract=contract.renewals())
    print(V)
    contract = Contract(premium=30*12, renewal_premium=0.05)
    V1 = life.gross_future_loss(a=a, contract=contract.renewals())
    print(V, V1, V+V1)
    print()
     

    print("SOA Question 6.25:  (C) 12330")
    life = SULT()
    woolhouse = Woolhouse(m=12, life=life)
    benefits = woolhouse.deferred_annuity(55, u=10, b=1000 * 12)
    expenses = life.whole_life_annuity(55, b=300)
    payments = life.temporary_annuity(55, t=10)
    print(benefits + expenses, payments)
    def fun(P):
        return life.gross_future_loss(A=benefits + expenses, a=payments,
                                      contract=Contract(premium=P))
    P = life.solve(fun, target=-800, grid=[12110, 12550])
    print(P)
    print()
     

    print("SOA Question 6.15:  (B) 1.002")
    life = Recursion().set_interest(i=0.05).set_a(3.4611, x=0)
    A = life.insurance_twin(3.4611)
    udd = UDD(m=4, life=life)
    a1 = udd.whole_life_annuity(x=x)
    woolhouse = Woolhouse(m=4, life=life)
    a2 = woolhouse.whole_life_annuity(x=x)
    print(life.gross_premium(a=a1, A=A)/life.gross_premium(a=a2, A=A))
    print()
     
    print("SOA Question 5.7:  (C) 17376.7")
    life = Recursion().set_interest(i=0.04)
    life.set_A(0.188, x=35)
    life.set_A(0.498, x=65)
    life.set_p(0.883, x=35, t=30)
    mthly = Woolhouse(m=2, life=life, three_term=False)
    print(mthly.temporary_annuity(35, t=30))
    print(1000 * mthly.temporary_annuity(35, t=30))
    print()



"""Interest Theory - Applies interest rate formulas

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
import math
import numpy as np
import pandas as pd
from typing import Callable
from actuarialmath.actuarial import Actuarial

class Interest(Actuarial):
    """Convert between nominal, discount, continously-compounded and 1/m'thly rates,
    and compute value of annuity certain
    
    Args:
      i : assumed annual interest rate
      d : or assumed discount rate
      v : or assumed discount factor
      delta : or assumed continuously compounded interest rate
      v_t : or assumed discount rate as a function of time
      i_m : or assumed monthly interest rate
      d_m : or assumed monthly discount rate
      m : m'thly frequency, if i_m or d_m are given

    Attributes:
      i (float) : interest rate
      d (float) : discount rate
      delta (float) :  continuously compounded interest rate
      v (float) : discount factor
      v_t (Callable) : discount factor as a function of time
    """

    def __init__(self, i: float = -1., delta: float = -1., d: float = -1., 
                 v: float = -1., i_m: float = -1., d_m: float = -1., m: int = 0,
                 v_t: Callable[[float], float] | None = None):
        if i_m >= 0:  # given interest rate mthly compounded
            i = self.mthly(m=m, i_m=i_m)
        if d_m >= 0:  # given discount rate mthly compounded
            d = self.mthly(m=m, d_m=d_m)
        if v_t is None:
            if delta >= 0:     # given continously-compounded rate
                self.i = math.exp(delta) - 1
            elif d >= 0:       # given annual discount rate
                self.i = d / (1 - d)
            elif v >= 0 :      # given annual discount factor
                self.i = (1 / v) - 1
            elif i >= 0:
                self.i = i
            else:              # given annual interest rate
                raise Exception("non-negative interest rate not given")
            self.v_t = lambda t: self.v**t 
            self.v = 1 / (1 + self.i)         # store discount factor
            self.d = self.i / (1 + self.i)    # store discount rate
            self.delta = math.log(1 + self.i) # store continuous rate
        else:   # given discount function
            assert callable(v_t), "v_t must be a callable discount function"
            assert v_t(0) == 1, "v_t(t=0) must equal 1"
            self.v_t = v_t
            #self.i = (1 / v_t(1)) - 1
            self.v = self.d = self.i = self.delta = None

    def annuity(self, t: int = -1, m: int = 1, due: bool = True) -> float:
        """Compute value of the annuity certain factor

        Args:
          t : number of years of payments
          m : m'thly frequency of payments (0 for continuous payments)
          due : whether annuity due (True) or immediate (False)
        """
        v_t = 0 if t < 0 else self.v**t   # is t finite
        assert m >= 0, "mthly frequency must be non-negative"
        if m == 0:  # if continuous
            return (1 - v_t) / self.delta
        elif due:   # if annuity due
            return (1 - v_t) / self.mthly(m=m, d=self.d)
        else:       # if annuity immediate
            return (1 - v_t) / self.mthly(m=m, i=self.i)

    @staticmethod
    def mthly(m: int = 0, i: float = -1, d: float = -1,
            i_m: float = -1, d_m: float = -1) -> float:
        """Convert to or from m'thly interest rates

        Args:
          m : m'thly frequency
          i : an annual-pay interest rate, to convert to m'thly
          d : or annual-pay discount rate, to convert to m'thly
          i_m : or m'thly interest rate, to convert to annual pay
          d_m : or m'thly discount rate, to convert to annual pay

        Examples:
          >>> i = Interest.mthly(i_m=0.05, m=12)
          >>> print("Convert mthly to annual-pay:", i)
          >>> print("Convert annual-pay to mthly:", Interest.mthly(i=i, m=12))
        """
        assert m >= 0, "mthly frequency must be non-negative"
        if i > 0:
            return m * ((1 + i)**(1 / m) - 1) if m else math.log(1+i)
        elif d > 0:
            return m * (1 - (1 - d)**(1 / m)) if m else -math.log(1-d)
        elif i_m > 0:
            return (1 + (i_m / m))**m - 1
        elif d_m > 0:
            return 1 - (1 - (d_m / m))**m
        else:
            raise Exception("no interest rate given to mthly")

    @staticmethod
    def double_force(i: float = -1., delta: float = -1., d: float = -1., 
                        v: float = -1.):
        """Double the force of interest

        Args:
          i : interest rate to double force of interest
          d : or discount rate
          v : or discount factor
          delta : or continuous rate

        Returns:
          interest rate, of same form as input rate, after doubling force of interest

        Examples:
          >>> print("Double force of interest of i =", Interest.double_force(i=0.05))
          >>> print("Double force of interest of d =", Interest.double_force(d=0.05))
        """
        if delta >= 0:
            return 2 * delta
        elif v >= 0:
            return v**2
        elif d >= 0:
            return 2 * d - (d**2)
        elif i >= 0:
            return 2 * i + (i**2)
        else:
            raise Exception("no interest rate for double_force")

if __name__ == "__main__":
    print("SOA Question 3.10:  (C) 0.86")
    interest = Interest(v=0.75)
    L = 35 * interest.annuity(t=4, due=False) + 75 * interest.v_t(t=5)
    interest = Interest(v=0.5)
    R = 15 * interest.annuity(t=4, due=False) + 25 * interest.v_t(t=5)
    ans = L / (L + R)
    print(ans)
    
    print("Double the force of interest")
    i = 0.05
    i2 = Interest.double_force(i=i)
    print(i2)

    print("Convert interest to discount rate")
    d2 = Interest(i=i2).d
    print(d2)

    print("Convert mthly to annual-pay:")
    i = Interest.mthly(i_m=0.05, m=12)
    print(i)

    print("Convert annual-pay to mthly:")
    i_m = Interest.mthly(i=i, m=12)
    print(i_m)

    



"""1/Mthly - Calculates m'thly-pay insurance and annuities

MIT License. Copyright 2022-2023 Terence Lim
"""
from typing import Callable
import math
import pandas as pd
from actuarialmath.annuity import Annuity
from actuarialmath.life import Actuarial

class Mthly(Actuarial):
    """Compute 1/M'thly insurance and annuities

    Args:
      m : number of payments per year
      life : original survival and life contingent functions
    """
    _methods = ['v_m', 'p_m', 'q_m', 'Z_m', 'E_x', 'A_x',
                 'whole_life_insurance', 'term_insurance', 'deferred_insurance',
                 'endowment_insurance', 'immediate_annuity', 'insurance_twin',
                 'annuity_twin', 'annuity_variance', 'whole_life_annuity',
                 'temporary_annuity', 'deferred_annuity', 'immediate_annuity']

    def __init__(self, m: int, life: Annuity):
        self.life = life
        self.m = max(0, m)
    
    def v_m(self, k: int) -> float:
        """Compute discount rate compounded over k m'thly periods

        Args:
          k : number of m'thly periods to compound
        """
        return self.life.interest.v_t(k / self.m)

    def q_m(self, x: int, s_m: int = 0, t_m: int = 1, u_m: int = 0) -> float:
        """Compute deferred mortality over m'thly periods

        Args:
          x : year of selection
          s_m : number of m'thly periods after selection
          u_m : survive number of m'thly periods , then
          t_m : dies within number of m'thly periods
        """
        sr = s_m / self.m
        s = math.floor(sr)
        r = sr - s
        q = self.life.q_r(x, s=s, r=r, t=t_m/self.m, u=u_m/self.m)
        return q

    def p_m(self, x: int, s_m: int = 0, t_m: int = 1) -> float:
        """Compute survival probability over m'thly periods

        Args:
          x : year of selection
          s_m : number of m'thly periods after selection
          t_m : survives number of m'thly periods
        """
        sr = s_m / self.m
        s = math.floor(sr)
        r = sr - s
        return self.life.p_r(x, s=s, r=r, t=t_m/self.m)

    def E_x(self, x: int, s: int = 0, t: int = 1, moment: int = 1,
            endowment: int = 1) -> float:
        """Compute pure endowment factor

        Args:
          x : year of selection
          s : years after selection
          t : term length in years
          moment : return first or second moment
          endowment : endowment amount
        """
        assert moment > 0
        return self.life.E_x(x, s=s, t=t, moment=moment) * endowment**moment

    def Z_m(self, x: int, s: int = 0, t: int = 1, 
            benefit: Callable = lambda x,t: 1, moment: int = 1):
        """Return PV of insurance r.v. Z and probability of death at mthly intervals

        Args:
          x : year of selection
          s : years after selection
          t : year of death
          benefit : amount of benefit by year and age selected
          moment : return first or second moment
        
        Returns:
          DataFrame, indexed by mthly period, with column names ['Z', 'p']
        """
        Z = [(benefit(x+s, k/self.m) * self.v_m(k+1))**moment 
             for k in range(t * self.m)]
        q = [self.q_m(x, s_m=s*self.m, u_m=k) for k in range (t*self.m)]
        return pd.DataFrame.from_dict(dict(m=range(1, self.m*t + 1), Z=Z, q=q))\
                           .set_index('m')

    def A_x(self, x: int, s: int = 0, t: int = 1, u: int = 0, 
            benefit: Callable = lambda x,t: 1, moment: int = 1) -> float:
        """Compute insurance factor with m'thly benefits

        Args:
          x : year of selection
          s : years after selection
          u : years deferred
          t : term of insurance in years
          benefit : amount of benefit by year and age selected
          moment : return first or second moment
        """
        assert moment in [1, 2]
        t = self.max_term(x+s, t)
        if self.m > 0:
            A = sum([(benefit(x+s, k/self.m) * self.v_m(k+1))**moment 
                     * self.q_m(x, s_m=s*self.m, u_m=k) 
                     for k in range((t+u) * self.m)])
        else:
            Z = lambda t: ((benefit(x+s, t+u) * self.life.v_t(t+u))**moment
                            * self.life.f(x, s, t+u))
            A = self.life.integrate(Z, 0, t)
        return A

    def whole_life_insurance(self, x: int, s: int = 0, moment: int = 1, 
                             b: int = 1) -> float:
        """Whole life insurance: A_x

        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          moment : compute first or second moment
        """
        assert moment in [1, 2, Annuity.VARIANCE]
        if moment == Annuity.VARIANCE:
            A2 = self.whole_life_insurance(x, s=s, moment=2)
            A1 = self.whole_life_insurance(x, s=s)
            return self.life.insurance_variance(A2=A2, A1=A1, b=b)
        return sum(self.A_x(x, s=s, b=b, moment=moment))


    def term_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                       moment: int = 1) -> float:
        """Term life insurance: A_x:t^1

        Args:
          x : year of selection
          s : years after selection
          t : term of insurance in years
          b : amount of benefit
          moment : return first or second moment
        """
        assert moment in [1, 2, Annuity.VARIANCE]
        if moment == Annuity.VARIANCE:
            A2 = self.term_insurance(x, s=s, t=t, moment=2)
            A1 = self.term_insurance(x, s=s, t=t)
            return self.life.insurance_variance(A2=A2, A1=A1, b=b)
        A = self.whole_life_insurance(x, s=s, b=b, moment=moment)
        if t < 0 or self.life.max_term(x+s, t) < t:
            return A
        E = self.E_x(x, s=s, t=t, moment=moment)
        A -= E * self.whole_life_insurance(x, s=s+t, b=b, moment=moment)
        return A

    def deferred_insurance(self, x: int, s: int = 0, n: int = 0, b: int = 1, 
                           t: int = Annuity._WHOLE, moment: int = 1) -> float:
        """Deferred insurance n|_A_x:t^1 = discounted whole life

        Args:
          x : year of selection
          s : years after selection
          u : years to defer
          t : term of insurance in years
          b : amount of benefit
          moment : return first or second moment
        """
        if self.life.max_term(x+s, n) < n:
            return 0.
        if moment == self.VARIANCE:
            A2 = self.deferred_insurance(x, s=s, t=t, n=n, moment=2)
            A1 = self.deferred_insurance(x, s=s, t=t, n=n)
            return self.life.insurance_variance(A2=A2, A1=A1, b=b)
        E = self.E_x(x, s=s, t=n, moment=moment)
        A = self.term_insurance(x, s=s+n, t=t, b=b, moment=moment)
        return E * A   # discount insurance by moment*force of interest

    def endowment_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                            endowment: int = -1, moment: int = 1) -> float:
        """Endowment insurance: A_x:t = term insurance + pure endowment

        Args:
          x : year of selection
          s : years after selection
          t : term of insurance in years
          b : amount of benefit
          endowment : amount of endowment
          moment : return first or second moment
        """
        if moment == self.VARIANCE:
            A2 = self.endowment_insurance(x, s=s, t=t, endowment=endowment, 
                                          b=b, moment=2)
            A1 = self.endowment_insurance(x, s=s, t=t, endowment=endowment, 
                                          b=b)
            return self.life.insurance_variance(A2=A2, A1=A1, b=b)
        E = self.E_x(x, s=s, t=t, moment=moment)
        A = self.term_insurance(x, s=s, t=t, b=b, moment=moment)
        return A + E * (b if endowment < 0 else endowment)**moment

    def insurance_twin(self, a: float) -> float:
        """Return insurance twin of m'thly annuity

        Args:
          a : twin annuity factor
        """
        d = self.life.interest.d
        d_m = self.life.interest.mthly(m=self.m, d=d)        
        return (1 - d_m * a)

    def annuity_twin(self, A: float) -> float:
        """Return value of annuity twin of m'thly insurance

        Args:
          A : amount of m'thly insurance
        """
        d = self.life.interest.d
        d_m = self.life.interest.mthly(m=self.m, d=d)        
        return (1-A) / d_m

    def annuity_variance(self, A2: float, A1: float, b: float = 1) -> float:
        """Variance of m'thly annuity from m'thly insurance moments

        Args:
          A2 : double force of interest of m'thly insurance
          A1 : first moment of m'thly insurance
          b : amount of benefit
        """
        num = self.life.insurance_variance(A2=A2, A1=A1, b=b)
        den = self.life.interest.mthly(m=self.m, d=self.life.interest.d)
        return num / den**2

    def whole_life_annuity(self, x: int, s: int = 0, b: int = 1, 
                           variance: bool = False) -> float:
        """Whole life m'thly annuity: a_x

        Args:
          x : year of selection
          s : years after selection
          b : amount of benefit
          variance : return first moment (False) or variance (True)
        """
        if variance:  # short cut for variance of whole life
            A1 = self.whole_life_insurance(x, s=s, moment=1)
            A2 = self.whole_life_insurance(x, s=s, moment=2)
            return self.annuity_variance(A2=A2, A1=A1, b=b)
        return b * (1 - self.whole_life_insurance(x, s=s)) / self.d
            
    def temporary_annuity(self, x: int, s: int = 0, t: int = Annuity._WHOLE, 
                          b: int = 1, variance: bool = False) -> float:
        """Temporary m'thly life annuity: a_x:t

        Args:
          x : year of selection
          s : years after selection
          t : term of annuity in years
          b : amount of benefit
          variance : return first moment (False) or variance (True)
        """
        if variance:  # short cut for variance of temporary life annuity
            A1 = self.term_insurance(x, s=s, t=t)
            A2 = self.term_insurance(x, s=s, t=t, moment=2)
            return self.annuity_variance(A2=A2, A1=A1, b=b)

        # difference of whole life on (x) and deferred whole life on (x+t)
        a = self.whole_life_annuity(x, s=s, b=b)
        if t < 0 or self.max_term(x+s, t) < t:
            return a
        a_t = self.whole_life_annuity(x, s=s+t, b=b)
        return a - (a_t * self.E_x(x, s=s, t=t))

    def deferred_annuity(self, x: int, s: int = 0, u: int = 0, 
                         t: int = Annuity._WHOLE, b: int = 1) -> float:
        """Deferred m'thly life annuity due n|t_a_x =  n+t_a_x - n_a_x

        Args:
          x : year of selection
          s : years after selection
          u : years of deferral
          t : term of annuity in years
          b : amount of benefit
        """
        if self.life.max_term(x+s, u) < u:
            return 0.
        return self.E_x(x, s=s, t=u)*self.temporary_annuity(x, s=s+u, t=t, b=b)

    def immediate_annuity(self, x: int, s: int = 0, t: int = Annuity._WHOLE, 
                          b: int = 1) -> float:
        """Immediate m'thly annuity

        Args:
          x : year of selection
          s : years after selection
          t : term of annuity in years
          b : amount of benefit
        """
        a = self.temporary_annuity(x, s=s, t=t)
        if self.m > 0:
            return (a - ((1 - self.E_x(x, s=s, t=t)) / self.m)) * b
        else:
            return a

if __name__ == "__main__":
    from actuarialmath.lifetable import LifeTable

    print("SOA Question 6.4:  (E) 1893.9")
    mthly = Mthly(m=12, life=Annuity().set_interest(i=0.06))
    A1, A2 = 0.4075, 0.2105
    mean = mthly.annuity_twin(A1)*15*12
    var = mthly.annuity_variance(A1=A1, A2=A2, b=15 * 12)
    S = Annuity.portfolio_percentile(mean=mean, variance=var, prob=.9, N=200)
    print(S / 200)
    print()
    
    print("SOA Question 4.2:  (D) 0.18")
    life = LifeTable(udd=False).set_table(q={0: 0.16, 1: 0.23})\
                               .set_interest(i_m=.18, m=2)
    mthly = Mthly(m=2, life=life)
    Z = mthly.Z_m(0, t=2, benefit=lambda x,t: 300000 + t*30000*2)
    print(Z)
    print(Z[Z['Z'] >= 277000].iloc[:, -1].sum())
    print()



"""Extra Risk - Adjusts force of mortality, age rating or mortality rate

MIT License. Copyright 2022-2023 Terence Lim
"""
from actuarialmath.survival import Survival
from actuarialmath.actuarial import Actuarial
from typing import Dict
import math

class ExtraRisk(Actuarial):
    """Adjust mortality by extra risk

    Args:
      life : original survival and mortality rates
      extra : amount of extra risk to adjust
      risk : adjust by {"ADD_FORCE" "MULTIPLY_FORCE" "ADD_AGE" "MULTIPLY_RATE"}
    """
    risks = ["ADD_FORCE", "MULTIPLY_FORCE", "ADD_AGE", "MULTIPLY_RATE"]

    def __init__(self, life: Survival, risk: str = "",
                 extra: float = 0.) -> "ExtraRisk":
        """Specify type and amount of mortality adjustment to apply"""
        assert not risk or risk in self.risks, "risk must be one of " + str(risks)
        assert extra >= 0, "amount of extra risk must be non-negative"
        self.life = life
        self.extra_ = extra
        self.risk_ = risk

    def __getitem__(self, col: str) -> Dict[int, float]:
        """Returns survival function values adjusted by extra risk

        Args:
          col : {'p', 'q'} for one-year survival or mortality function values

        Returns:
          dict of age and survival function values adjusted by extract risk
        """
        f = {'q': self.q_x, 'p': self.p_x}[col[0]]
        return {x: f(x) for x in range(self.life._MINAGE, self.life._MAXAGE+1)}

    def p_x(self, x: int, s: int = 0) -> float:
        """Return p_[x]+s after adding or multiplying force of mortality
    
        Args:
          x : age of selection
          s : years after selection
        """
        if self.risk_ in ["MULTIPLY_RATE"]:
            return 1 - self.q_x(x, s=s)
        if self.risk_ in ["ADD_AGE"]:
            return self.life.p_x(x + self.extra_, s=s)    
        p = self.life.p_x(x, s=s)
        if self.risk_ in ["MULTIPLY_FORCE"]:
            p = p**self.extra_
        if self.risk_ in ["ADD_FORCE"]:
            p *= math.exp(-self.extra_)
        return p

    def q_x(self, x: int, s: int = 0) -> float:
        """Return q_[x]+s after adding age rating or multipliying mortality rate
    
        Args:
          x : age of selection
          s : years after selection
        """
        if self.risk_ in ["ADD_FORCE", "MULTIPLY_FORCE"]:
            return 1 - self.p_x(x, s=s)
        if self.risk_ in ["ADD_AGE"]:
            return self.life.q_x(x + self.extra_, s=s)
        if self.risk_ in ["MULTIPLY_RATE"]:
            return self.extra_ * self.life.q_x(x, s=s)

if __name__ == "__main__":
    from actuarialmath.selectlife import SelectLife
    from actuarialmath.sult import SULT
    
    print("SOA Question 5.5: (A) 1699.6")
    life = SULT()
    extra = ExtraRisk(life=life, extra=0.05, risk="ADD_FORCE")
    select = SelectLife(periods=1)\
        .set_interest(i=.05)\
        .set_select(s=0, age_selected=True, q=extra['q'])\
        .set_select(s=1, age_selected=False, a=life['a'])\
        .fill_table()
    print(100*select['a'][45][0])
    print()

    print("SOA Question 4.19:  (B) 59050")
    life = SULT()
    extra = ExtraRisk(life=life, extra=0.8, risk="MULTIPLY_RATE")
    select = SelectLife(periods=1)\
        .set_interest(i=.05)\
        .set_select(s=0, age_selected=True, q=extra['q'])\
        .set_select(s=1, age_selected=False, q=life['q'])\
        .fill_table()
    print(100000*select.whole_life_insurance(80, s=0))
    print()
    
    print("Other usage examples")
    life = SULT()
    extra = ExtraRisk(life=life, extra=2, risk="MULTIPLY_FORCE")
    print(life.p_x(45), extra.p_x(45))



"""Reserves - Computes recursive, interim and modified reserves

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Callable, Dict, Any
from actuarialmath.policyvalues import PolicyValues, Contract

class Reserves(PolicyValues):
    """Compute recursive, interim and modified reserves"""
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._reserves = {'V': {}}
        self.T = 0

    #
    # Set up reserves table for recursion
    #
    def set_reserves(self, T: int = 0, endowment: int | float= 0,
                     V: Dict[int, float] | None = None) -> "Reserves":
        """Set values of the reserves table and the endowment benefit amount

        Args:
          T : max term of policy
          V : reserve values, keyed by time t
          endowment : endowment benefit amount
        """
        if T:
            self.T = T
        if V:
            self._reserves['V'].update(V)
            self.T = max(len(V) - 1, self.T)
        self._reserves['V'][0] = 0  # initial reserve is 0 by equivalence
        self._reserves['V'][self.T] = endowment  # n_V is 0 or endowment
        return self

    def fill_reserves(self, x: int, s: int = 0, reserve_benefit: bool = False,
                      contract: Contract | None = None) -> "Reserves":
        """Iteratively fill in missing values in reserves table

        Args:
          x : age selected
          s : starting from s years after selection
          reserve_benefit : whether benefit includes value of reserves
          contract : policy contract terms and expenses
        """
        contract = contract or Contract()        
        for _ in range(2):
            for t in range(self.T + 1):
                if self._reserves['V'].get(t, None) is not None:
                    continue
                if t == contract.T:
                    v = self.t_V(x=x, s=s, t=t, premium=0, benefit=lambda t: 0, 
                                 per_policy = -contract.endowment)
                elif t == 1:
                    v = self.t_V(x=x, s=s, t=t, premium=contract.premium, 
                                 benefit=lambda t: contract.benefit,
                                 reserve_benefit=reserve_benefit,
                                 per_premium=contract.initial_premium, 
                                 per_policy=contract.initial_policy)
                elif t == 0:
                    v = 0
                else:
                    v = self.t_V(x=x, s=s, t=t, premium=contract.premium, 
                                 benefit=lambda t: contract.benefit,
                                 reserve_benefit=reserve_benefit,
                                 per_premium=contract.renewal_premium, 
                                 per_policy=contract.renewal_policy)
                if v is not None:
                    self._reserves['V'][t] = v
        return self

    def V_plot(self, ax: Any = None, color: str = 'r', title: str = ''):
        """Plot values from reserves tables

        Args:
          title : title to display
          color : color to plot curve
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        y = [self._reserves['V'].get(t, None) for t in range(self.T + 1)]
        ax.plot(list(range(self.T + 1)), y, '.', color=color)
        ax.set_title(title)
        ax.set_ylabel(f"$_tV$", color=color)
        ax.set_xlabel(f"t")

    def V_t(self):
        """Returns reserves table as a DataFrame"""
        return pd.DataFrame(self._reserves)\
                 .rename_axis('t')\
                 .rename(columns={'V':'V_t'})

    #
    # Reserves recursion
    #
    def t_V_backward(self, x: int, s: int = 0, t: int = 0, premium: float = 0, 
                     benefit: Callable = lambda t: 1, 
                     per_premium: float = 0, per_policy: float = 0,
                     reserve_benefit: bool = False) -> float | None:
        """Backward recursion (with optional reserve benefit)

        Args:
          x : age selected
          s : starting s years after selection
          t : year of reserve to solve
          benefit : benefit amount at t+1
          premium : amount of premium paid just after t
          per_premium : expense per $ premium
          per_policy : expense per policy
          reserve_benefit : whether reserve value at t+1 included in benefit
        """
        if t+1 not in self._reserves['V']:
            return None
        V = self._reserves['V'][t+1]
        b = benefit(t+1) + V * reserve_benefit   # total death benefit 
        if V == b:   # special case if death benefit == forward reserve
            V = b
        else:
            if V:
                V *= self.p_x(x=x, s=s+t)
            if b:
                V += self.q_x(x=x, s=s+t) * b
        V = V * self.interest.v - (premium*(1 - per_premium) - per_policy)
        return V

    def t_V_forward(self, x: int, s: int = 0, t: int = 0, premium: float = 0,
                    benefit: Callable = lambda t: 1, 
                    per_premium: float = 0, per_policy: float = 0,
                    reserve_benefit: bool = False) -> float | None:
        """Forward recursion (with optional reserve benefit)

        Args:
          x : age selected
          s : starting s years after selection
          t : year of reserve to solve
          benefit : benefit amount at t
          premium : amount of premium paid just after t-1
          per_premium : expense per $ premium
          per_policy : expense per policy
          reserve_benefit : whether reserve value at t included in benefit
        """
        if t-1 not in self._reserves['V']:
            return None
        V = self._reserves['V'][t-1]
        V = (V + premium*(1 - per_premium) - per_policy) / self.interest.v
        if benefit(t):
            V -= self.q_x(x=x, s=s+t-1) * benefit(t)
        if not reserve_benefit:
            V /= self.p_x(x=x, s=s+t-1)
        return V

    def t_V(self, x: int, s: int = 0, t: int = 0,
            premium: float = 0, benefit: Callable = lambda t: 1, 
            reserve_benefit: bool = False,
            per_premium: float = 0, per_policy: float = 0) -> float | None:
        """Solve year-t reserves by forward or backward recursion

        Args:
          x : age selected
          s : starting s years after selection
          t : year of reserve to solve
          benefit : benefit amount
          premium : amount of premium
          per_premium : expense per $ premium
          per_policy : expense per policy
          reserve_benefit : whether reserve value included in benefit
        """
        if t in self._reserves['V']:   # already solved in reserves table
            return self._reserves['V'][t]
        V = self.t_V_forward(x=x, s=s, t=t, premium=premium, 
                             benefit=benefit,
                             reserve_benefit=reserve_benefit, 
                             per_premium=per_premium, 
                             per_policy=per_policy)
        if V is not None:
            return V
        V = self.t_V_backward(x=x, s=s, t=t, premium=premium, 
                              benefit=benefit,
                              reserve_benefit=reserve_benefit, 
                              per_premium=per_premium, 
                              per_policy=per_policy)
        if V is not None:
            return V

    #
    # Interim reserves
    #
    def r_V_backward(self, x: int, s: int = 0, r: float = 0,
                   benefit: int = 1) -> float | None:
        """Backward recursion for interim reserves

        Args:
          x : age of selection
          s : years after selection
          r : solve for interim reserve at fractional year x+s+r
          benefit : benefit amount in year x+s+1
        """
        s = int(s + r)
        r = r - math.floor(r)
        if s+1 not in self._reserves['V']:        # forward recursion
            return None
        V = self._reserves['V'][s+1]
        if V:
            V *= self.p_r(x, s=s, r=r, t=1-r)
        if benefit:
            V += self.q_r(x, s=s, r=r, t=1-r) * benefit
        V = V * self.interest.v_t(1-r)
        return V

    def r_V_forward(self, x: int, s: int = 0, r: float = 0,
                    premium: float = 0, benefit: int = 1) -> float | None:
        """Forward recursion for interim reserves

        Args:
          x : age of selection
          s : years after selection
          r : solve for interim reserve at fractional year x+s+r
          benefit : benefit amount in year x+s+1
          premium : premium amount just after year x+s
        """
        s = int(s + r)
        r = r - math.floor(r)
        if s not in self._reserves['V']:
            return None
        V = self._reserves['V'][s]
        V = (V + premium) / self.interest.v_t(r)
        if benefit:
            V -= self.q_r(x, s=s, t=r) * benefit * self.interest.v_t(1-r)
        V /= self.p_r(x, s=s, t=r)
        return V

    #
    # Full Preliminary Term (FPT) modified reserves
    #
    def FPT_premium(self, x: int, s: int = 0, n: int = PolicyValues._WHOLE, 
                    b: int = 1, first: bool = False) -> float:
        """Initial or renewal Full Preliminary Term premiums

        Args:
          x : age of selection
          s : years after selection
          n : term of insurance
          b : benefit amount in year x+s+1
          first : calculate year 1 (True) or year 2+ (False) FPT premium
        """
        if first:
            return self.net_premium(x, s=s, b=b, t=1)
        else:
            return self.net_premium(x, s=s+1, b=b, t=self.add_term(n, -1))

    def FPT_policy_value(self, x: int, s: int = 0, t: int = 0, b: int = 1,
                         n: int = PolicyValues._WHOLE,
                         endowment: int = 0, discrete: bool = True) -> float:
        """Compute Full Preliminary Term policy value at time t

        Args:
          x : age of selection
          s : years after selection
          n : term of insurance
          t : year of policy value to calculate
          b : benefit amount in year x+s+1
          endowment : endowment amount
          discrete : fully discrete (True) or continuous (False) insurance
        """
        if t in [0, 1]:  # FPT is 0 at t = 0 or 1
            return 0
        else:
            return self.net_policy_value(x, s=s+1, t=t-1, n=self.add_term(n,-1),
                                         b=b, endowment=endowment,
                                         discrete=discrete)


if __name__ == "__main__":
    from actuarialmath.sult import SULT
    from actuarialmath.policyvalues import Contract
    life = SULT()
    x, T, b = 50, 20, 500000    # $500K 20-year term insurance for (50)
    P = life.net_premium(x=x, t=T, b=b)
    life.set_reserves(T=T)\
        .fill_reserves(x=x, contract=Contract(premium=P, benefit=b))
    life.V_plot(title=f"Reserves for ${b} {T}-year term insurance issued to ({x})")

if False:
    print("SOA Question 7.31:  (E) 0.310")
    x = 0
    life = Reserves().set_reserves(T=3)
    print(life._reserves)
    G = 368.05
    def fun(P):  # solve net premium from expense reserve equation
        return life.t_V(x=x, t=2, premium=G-P, benefit=lambda t: 0,
                        per_policy=5+.08*G)
    P = life.solve(fun, target=-23.64, grid=[.29, .31]) / 1000
    print(P)
    print()

    from actuarialmath.sult import SULT
    print("SOA Question 7.13: (A) 180")
    life = SULT()
    V = life.FPT_policy_value(40, t=10, n=30, endowment=1000, b=1000)
    print(V)
    print()



"""Select and Ultimate Life Table -- Loads and calculates select life tables

MIT License. Copyright 2022-2023 Terence Lim
"""
from typing import Dict, List
import math
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from actuarialmath.lifetable import LifeTable

class SelectLife(LifeTable):
    """Calculate select life table, and iteratively fill in missing values

    Args:
      periods : number of select period years
      verbose : whether to echo update steps

    Notes:
      6 types of information can be loaded and calculated in the select table:

      - 'q' : probability [x]+s dies in one year
      - 'l' : number of lives aged [x]+s
      - 'd' : number of deaths of age [x]+s
      - 'A' : whole life insurance
      - 'a' : whole life annuity
      - 'e' : expected future curtate lifetime of [x]+s
    """

    def __init__(self, periods: int = 0, udd: bool = True,
                 verbose: bool = False, **kwargs):
        super().__init__(udd=udd, **kwargs)
        self._select = {'l':{}, 'd':{}, 'q':{}, 'e':{}, 'A':{}, 'a':{}}
        self.periods_ = periods
        self._verbose = verbose

    def __getitem__(self, table: str) -> Dict[int, float]:
        """Returns values from a select and ultimate table

        Args:
          table :  may be {'l', 'q', 'e', 'd', 'a', 'A'}
        """
        return self._select[table[0]]

    def set_table(self, fill: bool = True,
                  l: Dict[int, List[float]] | None = None,
                  d: Dict[int, List[float]] | None = None, 
                  q: Dict[int, List[float]] | None = None,
                  A: Dict[int, List[float]] | None = None,
                  a: Dict[int, List[float]] | None = None, 
                  e: Dict[int, List[float]] | None = None) -> "SelectLife":
        """Update from table, every age has row for all select durations

        Args:
          q : probability [x]+s dies in one year
          l : number of lives aged [x]+s
          d : number of deaths of [x]+s
          A : whole life insurance, or
          a : whole life annuity, or 
          e : expected future lifetime of [x]+s
        """
        periods = self.periods_    # infer number of select years, and age range
        minage = self._MINAGE 
        maxage = self._MAXAGE
        
        for lbl, table in [('A',A), ('a',a), ('q',q), ('d',d), ('l',l), ('e',e)]:
            self._select[lbl] = {}
            if table:     # update given table cells
                for age, row in table.items():
                    periods = max(len(row) - 1, periods)
                    minage = age if minage < 0 else min(minage, age)
                    maxage = age if maxage < 0 else max(maxage, age)
                    self._select[lbl][age] = {k:v for k,v in enumerate(row)}

        self.periods_ = periods    # update number of select years, and age range
        self._MINAGE = minage
        self._MAXAGE = maxage

        if fill:         # iteratively fill missing table values
            self.fill_table()
        return self

    def set_select(self, s: int, age_selected: bool, fill: bool = False,
                   l: Dict[int, float] | None = None,
                   d: Dict[int, float] | None = None,
                   q: Dict[int, float] | None = None,
                   A: Dict[int, float] | None = None,
                   a: Dict[int, float] | None = None,
                   e: Dict[int, float] | None = None) -> "SelectLife":
        """Update a table column, for a particular duration s in the select period

        Args:
          s : column to populate - n is ultimate, 0..n-1 is year after select
          age_selected : is indexed by age selected or actual (False, default)
          q : probabilities [x]+s dies in next year, by age
          l : number of lives aged [x]+s, by age
          d : number of deaths of [x]+s, by age
          A : whole life insurance of [x]+s, by age
          a : whole life annuity of [x]+s, by age
          e : expected future lifetime of [x]+s, by age
        """
        def ifelse(x, y): return y if x is None else x
        
        s = self.periods_ if s < 0 else s
        tables = {
            'l': ifelse(l, {}),
            'q': ifelse(q, {}),
            'd': ifelse(e, {}),
            'e': ifelse(e, {}),
            'a': ifelse(a, {}),
            'A': ifelse(A, {})
        }
        for table, item in tables.items():
            if item:
                if table not in self._select:
                    self._select[table] = {}
                for age in item.keys():
                    x = age - s * (1 - age_selected)          # get true age index
                    if x not in self._select[table]:
                        self._select[table][x] = {}
                    self._select[table][x][s] = item[age]

                    if self._MINAGE < 0 or x < self._MINAGE:  # infer min age
                        self._MINAGE = x
                    if age > self._MAXAGE:                    # infer max age
                        self._MAXAGE = x
        if fill:
            self.fill_table()
        return self

    def get_sel(self, x: int, s: int, table: str) -> float | None:
        """Helper to read right across, and down if neccesary (when s > n)"""
        if s > self.periods_:
            x += (s - self.periods_)
            s = self.periods_
        if (x in self._select[table] and s in self._select[table][x] 
            and self._select[table][x][s] is not None):
            return self._select[table][x][s]  # in select

    def isin_sel(self, x: int, s: int, table: str) -> bool:
        """Helper to check if value not missing"""
        return self.get_sel(x, s, table) is not None

    def fill_table(self, radix: int = LifeTable._RADIX) -> "SelectLife":
        """Fills in missing table values (does not check for consistency)

        Args:
          radix : initial number of lives
        """
        
        def A_x(x: int, s: int) -> float | None:
            """Helper to apply backward and forward recursion for insurance A_x"""
            if self.isin_sel(x, s, 'A'):
                return self.get_sel(x, s, 'A')

            _x, _s = (x, s+1) if s < self.periods_ else (x+1, s)  # right or down
            if self.isin_sel(x, s, 'q') and self.isin_sel(_x, _s, 'A'):
                q = self.get_sel(x, s, 'q')  # A_x = (v q) + (v p A_x+1)
                return self.interest.v * (q + (1-q)*self.get_sel(_x, _s, 'A'))

            backwards = [(x, s-1)]   # to move backwards along row
            if s >= self.periods_:      # if in ultimate, can also move up column
                backwards += [(x-1, s)]
            for _x, _s in backwards:   # A_x+1 = (A_x - qv) / (p v)
                if self.isin_sel(_x, _s, 'q') and self.isin_sel(_x, _s, 'A'):
                    q = self.get_sel(_x, _s, 'q')
                    return ((self.get_sel(_x, _s, 'A') - q * self.interest.v)
                            / (self.interest.v * (1 - q)))

        def a_x(x: int, s: int) -> float | None:
            """Helper to apply backward and forward recursion for annuity a_x"""
            if self.isin_sel(x, s, 'a'):
                return self.get_sel(x, s, 'a')

            _x, _s = (x, s+1) if s < self.periods_ else (x+1, s)  # right or down
            if self.isin_sel(x, s, 'q') and self.isin_sel(_x, _s, 'a'):
                p = 1 - self.get_sel(x, s, 'q')  # a_x = 1 +  (v p a_x+1)
                return 1 + self.interest.v * p * self.get_sel(_x, _s, 'a')

            backwards = [(x, s-1)]  # to move backwards along row
            if s >= self.periods_:      # if in ultimate, can also move up column
                backwards += [(x-1, s)]
            for _x, _s in backwards:   # a_x+1 = (a_x - 1) / (p v)
                if self.isin_sel(_x, _s, 'q') and self.isin_sel(_x, _s, 'a'):
                    p = 1 - self.get_sel(_x, _s, 'q')
                    return (self.get_sel(_x, _s, 'a') - 1)/(p*self.interest.v)


        def l_x(x: int, s: int) -> float | None:
            """Helper to solve for number of lives aged [x]+s: l_[x]+s"""
            if self.isin_sel(x, s, 'l'):
                return self.get_sel(x, s, 'l')
            if self.isin_sel(x, s-1, 'l') and self.isin_sel(x, s-1, 'q'):
                return self.get_sel(x, s-1, 'l')*(1 - self.get_sel(x, s-1, 'q'))
            if self.isin_sel(x, s+1, 'l') and self.isin_sel(x, s, 'q'):
                return self.get_sel(x, s+1, 'l') / (1 - self.get_sel(x, s, 'q'))
            if self.isin_sel(x, s, 'd') and self.isin_sel(x, s, 'q'):
                return self.get_sel(x, s, 'd') / self.get_sel(x, s, 'q')

            backwards = [(x, s-1)]  # to move backwards along row
            if s >= self.periods_:     # if in ultimate, can also move up column
                backwards += [(x-1, s)]
            for _x, _s in backwards:   # l_x+1 = l_x * p_x
                if self.isin_sel(_x, _s, 'l') and self.isin_sel(_x, _s, 'q'):
                    p = 1 - self.get_sel(_x, _s, 'q')
                    return self.get_sel(_x, _s, 'l') * p

        def d_x(x: int, s: int) -> float | None:
            """Helper to solve for deaths at [x]+s: l_[x]+s - l_[x]+s+1"""
            if self.isin_sel(x, s, 'd'):
                return self.get_sel(x, s, 'd')
            if self.isin_sel(x, s+1, 'l') and self.isin_sel(x, s, 'l'):
                return self.get_sel(x, s, 'l') - self.get_sel(x, s+1, 'l')

        def q_x(x: int, s: int) -> float | None:
            """Helper to solve for one-year mortality [x]+s: q_[x]+s"""
            if self.isin_sel(x, s, 'q'):
                return self.get_sel(x, s, 'q')
            if self.isin_sel(x, s, 'd') and self.isin_sel(x, s, 'l'):
                return self.get_sel(x, s, 'd') / self.get_sel(x, s, 'l')
            if self.isin_sel(x, s, 'e') and self.isin_sel(x, s+1, 'e'):
                return 1 - (self.get_sel(x, s, 'e') 
                            / (1 + self.get_sel(x, s+1, 'e')))

        def e_x(x: int, s: int) -> float | None:
            """Helper to solve for expected kurtate lifetime: e_[x]+s"""
            if self.isin_sel(x, s, 'e'):
                return self.get_sel(x, s, 'e')

            _x, _s = (x, s+1) if s < self.periods_ else (x+1, s) # right or down
            if self.isin_sel(x, s, 'q') and self.isin_sel(_x, _s, 'e'):
                return ((1 - self.get_sel(x, s, 'q'))
                        * (1 + self.get_sel(_x, _s, 'e')))  # e_x = p(1 + e_x+1)

            backwards = [(x, s-1)]  # to move backwards along row
            if s >= self.periods_:     # if in ultimate, can also move up column
                backwards += [(x-1, s)]
            for _x, _s in backwards:   # e_x+1 = e_x/p_x - 1
                if self.isin_sel(_x, _s, 'e') and self.isin_sel(_x, _s, 'q'):
                    return (self.get_sel(_x, _s, 'e') 
                            / (1 - self.get_sel(_x, _s, 'q'))) - 1

        # Iterate a few times to impute select table values
        funs = {'l': l_x, 'A': A_x, 'a': a_x, 'q': q_x, 'd': d_x, 'e': e_x}
        curr = 0   # number of updates filled in
        for loop in range(2):   # loop second time if need to assume radix
            prev = curr - 1
            while curr != prev:        # continue while changes made
                prev = curr
                for tab, fun in funs.items():   # for each table
                    if tab not in self._select:
                        self._select[tab] = {}
                    for x in range(self._MINAGE, self._MAXAGE + 1):  # for each age
                        if x not in self._select[tab]:
                            self._select[tab][x] = {}
                        for s in range(self.periods_+1): # each year after select
                            if not self.isin_sel(x, s, tab):
                                val = fun(x, s)
                                if val is not None:
                                    self._select[tab][x][s] = val
                                    curr += 1
                                    if self._verbose:
                                        print(f"{curr} {tab}(x={x}, s={s}) = {val}")
            if 0 not in self._select['l'][self._MINAGE]: # arbitrary initial lives
                self._select['l'][self._MINAGE][0] = radix
        return self

    def A_x(self, x: int, s: int = 0, moment: int = 1, discrete: bool = True,
            **kwargs) -> float:
        """Returns insurance value computed from select table

        Args:
          x : age of selection
          s : years after selection
        """
        assert moment == 1 and discrete, "Must be discrete insurance"
        if self.isin_sel(x, s, 'A'):
            return self.get_sel(x, s, 'A')
        else:
            return super().A_x(x=x, s=s, moment=1, discrete=True, **kwargs)

    def a_x(self, x: int, s: int = 0, moment: int = 1, discrete: bool = True,
            **kwargs) -> float:
        """Returns annuity value computed from select table

        Args:
          x : age of selection
          s : years after selection
        """
        assert moment == 1 and discrete, "Must be discrete annuity"
        if self.isin_sel(x, s, 'a'):
            return self.get_sel(x, s, 'a')
        else:
            return super().A_x(x=x, s=s, moment=1, discrete=True, **kwargs)

    def l_x(self, x: int, s: int = 0) -> float:
        """Returns number of lives aged [x]+s computed from select table

        Args:
          x : age of selection
          s : years after selection
        """
        return self.get_sel(x, s, 'l')

    def p_x(self, x: int, s: int = 0, t: int = 1) -> float:
        """t_p_[x]+s by chain rule: prod(1_p_[x]+s+y) for y in range(t)

        Args:
          x : age of selection
          s : years after selection
          t : survives t years
        """
        return np.prod([1 - self.get_sel(x, s+y, 'q') for y in range(t)])

    def q_x(self, x: int, s: int = 0, t: int = 1, u: int = 0) -> float:
        """t|u_q_[x]+s = [x]+s survives u years, does not survive next t

        Args:
          x : age of selection
          s : years after selection
          u : survives u years, then
          t : dies within next t years
        """
        return (1. - self.p_x(x, s=s + u, t=t)) * self.p_x(x, s=s, t=u)

    def e_x(self, x: int, s: int = 0, t: int = LifeTable._WHOLE,
            curtate: int = True) -> float:
        """Returns expected life time computed from select table

        Args:
          x : age of selection
          s : years after selection
          t : limit of expected future lifetime
        """
        assert curtate, "Must be curtate lifetimes"
        if (self.isin_sel(x, s, 'e')):
            return self.get_sel(x, s, 'e')
        e = sum([self.p_x(x, s=s, t=k+1) for k in range(max(1, t))])
        return e

    def frame(self, table: str = 'l') -> pd.DataFrame:
        """Returns select and ultimate table values as a DataFrame

        Args:
          table : table to return, one of ['A', 'a', 'q', 'd', 'e', 'l']
        """
        return pd.DataFrame.from_dict(self._select[table[0]], orient='index')\
                           .sort_index(axis=0)\
                           .sort_index(axis=1)\
                           .rename_axis('Age')\
                           .rename_axis(table[0] + '_[x]+s:', axis=1)

if __name__ == "__main__":
    print("SOA Question 3.2:  (D) 14.7")
    e_curtate = SelectLife.e_approximate(e_complete=15)
    life = SelectLife(udd=True).set_table(l={65: [1000, None,],
                                             66: [955, None]},
                                          e={65: [e_curtate, None]},
                                          d={65: [40, None,],
                                             66: [45, None]})
    print(life.e_r(66))
    print(life.frame('e'))
    print()
    

    print("SOA Question 4.16:  (D) .1116")
    q = [.045, .050, .055, .060]
    q_ = {50+x: [0.7 * q[x] if x < 4 else None, 
                 0.8 * q[x+1] if x+1 < 4 else None, 
                 q[x+2] if x+2 < 4 else None] 
          for x in range(4)}
    life = SelectLife(verbose=True).set_table(q=q_).set_interest(i=.04)
    print(life.term_insurance(50, t=3))
    print()
    
    print("SOA Question 4.13:  (C) 350 ")
    life = SelectLife().set_interest(i=0.04)\
                           .set_table(q={65: [.08, .10, .12, .14],
                                         66: [.09, .11, .13, .15],
                                         67: [.10, .12, .14, .16],
                                         68: [.11, .13, .15, .17],
                                         69: [.12, .14, .16, .18]})
    print(life.deferred_insurance(65, t=2, u=2, b=2000))
    print()
    
    print("SOA Question 3.13:  (B) 1.6")
    life = SelectLife().set_table(l={55: [10000, 9493, 8533, 7664],
                                     56: [8547, 8028, 6889, 5630],
                                     57: [7011, 6443, 5395, 3904],
                                     58: [5853, 4846, 3548, 2210]},
                                  e={57: [None, None, None, 1]})
    print(life.e_r(58, s=2))
    print()


    print("SOA Question 3.12: (C) 0.055 ")
    life = SelectLife(udd=False).set_table(l={60: [10000, 9600, 8640, 7771],
                                              61: [8654, 8135, 6996, 5737],
                                              62: [7119, 6549, 5501, 4016],
                                              63: [5760, 4954, 3765, 2410]})
    print(life.q_r(60, s=1, t=3.5) - life.q_r(61, s=0, t=3.5))
    print()
    

    print("SOA Question 3.7: (b) 16.4")
    life = SelectLife().set_table(q={50: [.0050, .0063, .0080],
                                     51: [.0060, .0073, .0090],
                                     52: [.0070, .0083, .0100],
                                     53: [.0080, .0093, .0110]})
    print(1000*life.q_r(50, s=0, r=0.4, t=2.5))
    print()
    


    print("SOA Question 3.6:  (D) 5.85")
    life = SelectLife().set_table(q={60: [.09, .11, .13, .15],
                                     61: [.1, .12, .14, .16],
                                     62: [.11, .13, .15, .17],
                                     63: [.12, .14, .16, .18],
                                     64: [.13, .15, .17, .19]},
                                  e={61: [None, None, None, 5.1]})
    print(life.e_x(61))
    print()
    

    print("SOA Question 3.3:  (E) 1074")
    life = SelectLife().set_table(l={50: [99, 96, 93],
                                     51: [97, 93, 89],
                                     52: [93, 88, 83],
                                     53: [90, 84, 78]})
    print(10000*life.q_r(51, s=0, r=0.5, t=2.2))
    print()
    

    print("SOA Question 3.1:  (B) 117")
    life = SelectLife().set_table(l={60: [80000, 79000, 77000, 74000],
                                     61: [78000, 76000, 73000, 70000],
                                     62: [75000, 72000, 69000, 67000],
                                     63: [71000, 68000, 66000, 65000]})
    print(1000*life.q_r(60, s=0, r=0.75, t=3, u=2))
    print()
    

    print("Other usage examples")
    table={21: [.00120, .00150, .00170, .00180],
           22: [.00125, .00155, .00175, .00185],
           23: [.00130, .00160, .00180, .00195]}
    life = SelectLife(verbose=True).set_table(q=table)
    print(life.frame('l').round(1))
    print('--------------')
    print(life.frame('q').round(6))
    print('==============')
    print(life.p_x(21, 1, 4))  #0.99317
    
    



"""SULT - Loads and uses a standard ultimate life table

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
import numpy as np
import pandas as pd
from typing import Dict, Callable
from actuarialmath.lifetable import LifeTable

def faml_survival(x, t, A = 0.00022, B = 0.0000027, c = 1.124):
    """SULT default survival function from SOA's `Excel Workbook for FAM-L Tables`"""
    A, B, c = 0.00022, 0.0000027, 1.124
    return math.exp(-A*t - (B*c**x*(c**t - 1))/math.log(c))

class SULT(LifeTable):
    """Generates and uses a standard ultimate life table

    Args:
      i : interest rate
      radix : initial number of lives
      minage : minimum age
      maxage : maximum age
      S : survival function, default is Makeham with SOA FAM-L parameters
    """
    
    def __init__(self, i: float = 0.05, radix: int = 100000, 
                 S: Callable[[float, float], float] = faml_survival,
                 minage: int = 20, maxage: int = 130, **kwargs):
        """Construct SULT"""
        super().__init__(**kwargs)
        l = {t+minage: radix * S(minage, t) for t in range(1+maxage-minage)}
        self.set_interest(i=i).set_table(l=l, minage=minage, maxage=maxage)

    def __getitem__(self, col: str) -> Dict[int, float]:
        """Returns a column of the sult table

        Args:
          col : name of life table column to return
        """
        funs = {'q': (self.q_x, dict()),
                'p': (self.p_x, dict()),                
                'a': (self.whole_life_annuity, dict()),
                'A': (self.whole_life_insurance, dict())}
        assert col[0].lower() in funs, f"must be one of {list(funs.keys())}"
        f, args = funs[col[0]]
        return {x: f(x, **args) for x in range(self._MINAGE, self._MAXAGE)}

    def frame(self, minage: int = 20, maxage: int = 100):
        """Derive FAM-L exam table columns of SULT as a DataFrame

        Args:
          minage : first age to display row
          maxage : large age to display row
        """
        # specify methods and arguments for computing columns of FAM-L exam table
        funs = {'q_x': (self.q_x, dict()),
                'a_x': (self.whole_life_annuity, dict()),
                'A_x': (self.whole_life_insurance, dict()),
                '2A_x': (self.whole_life_insurance, dict(moment=2)),
                'a_x:10': (self.temporary_annuity, dict(t=10)),
                'A_x:10': (self.endowment_insurance, dict(t=10)),
                'a_x:20': (self.temporary_annuity, dict(t=20)),
                'A_x:20': (self.endowment_insurance, dict(t=20)),
                '5_E_x': (self.E_x, dict(t=5)),
                '10_E_x': (self.E_x, dict(t=10)),
                '20_E_x': (self.E_x, dict(t=20))}
        t = {col: {x: f(x, **args) for x in range(self._MINAGE, self._MAXAGE)}
             for col, (f, args) in funs.items()}
        tab = pd.DataFrame(dict(l_x=self._table['l'])).sort_index()
        tab = tab.join(pd.DataFrame.from_dict(t).set_index(tab.index[:-1]))
        for digits, col in zip([1, 6, 4, 5, 5, 4, 5, 4, 5, 5, 5, 5], tab.columns):
             tab[col] = tab[col].map(f"{{:.{digits}f}}".format)
        return tab.loc[minage:maxage]

if __name__ == "__main__":
    print("SOA Question 6.52:  (D) 50.80")
    sult = SULT()
    a = sult.temporary_annuity(45, t=10)
    other_cost = 10 * sult.deferred_annuity(45, u=10)
    P = sult.gross_premium(a=a, A=0, benefit=0, 
                           initial_premium=1.05, renewal_premium=0.05,
                           initial_policy=100 + other_cost, renewal_policy=20)
    print(a, P)
    print()
    

    print("SOA Question 6.47:  (D) 66400")
    sult = SULT()
    a = sult.temporary_annuity(70, t=10)
    A = sult.deferred_annuity(70, u=10)
    P = sult.gross_premium(a=a, A=A, benefit=100000, initial_premium=0.75,
                           renewal_premium=0.05)
    print(P)
    print()
    

    print("SOA Question 6.43:  (C) 170")
    sult = SULT()
    a = sult.temporary_annuity(30, t=5)
    A = sult.term_insurance(30, t=10)
    other_expenses = 4 * sult.deferred_annuity(30, u=5, t=5)
    P = sult.gross_premium(a=a, A=A, benefit=200000, initial_premium=0.35,
                           initial_policy=8 + other_expenses, renewal_policy=4,
                           renewal_premium=0.15)
    print(P)
    print()

    print("SOA Question 6.39:  (A) 29")
    sult = SULT()
    P40 = sult.premium_equivalence(sult.whole_life_insurance(40), b=1000)
    P80 = sult.premium_equivalence(sult.whole_life_insurance(80), b=1000)
    p40 = sult.p_x(40, t=10)
    p80 = sult.p_x(80, t=10)
    P = (P40 * p40 + P80 * p80) / (p80 + p40)
    print(P)
    print()
    

    print("SOA Question 6.37:  (D) 820")
    sult = SULT()
    benefits = sult.whole_life_insurance(35, b=50000 + 100)
    expenses = sult.immediate_annuity(35, b=100)
    a = sult.temporary_annuity(35, t=10)
    print(benefits, expenses, a)
    print((benefits + expenses) / a)
    print()
    

    print("SOA Question 6.35:  (D) 530")
    sult = SULT()
    A = sult.whole_life_insurance(35, b=100000)
    a = sult.whole_life_annuity(35)
    print(sult.gross_premium(a=a, A=A, initial_premium=.19, renewal_premium=.04))
    print()
    

    print("SOA Question 5.8: (C) 0.92118")
    sult = SULT()
    a = sult.certain_life_annuity(55, u=5)
    print(sult.p_x(55, t=math.floor(a)))
    print()
    

    print("SOA Question 5.3:  (C) 6.239")
    sult = SULT()
    t = 10.5
    print(t * sult.E_r(40, t=t))
    print()
    

    print("SOA Question 4.17:  (A) 1126.7")
    sult = SULT()
    median = sult.Z_t(48, prob=0.5, discrete=False)
    benefit = lambda x,t: 5000 if t < median else 10000
    print(sult.A_x(48, benefit=benefit))
    print()
    

    print("SOA Question 4.14:  (E) 390000    ")
    sult = SULT()
    p = sult.p_x(60, t=85-60)
    mean = sult.bernoulli(p)
    var = sult.bernoulli(p, variance=True)
    F = sult.portfolio_percentile(mean=mean, variance=var, prob=.86, N=400)
    print(F * 5000 * sult.interest.v_t(85-60))
    print()
    
    from actuarialmath.interest import Interest
    print("SOA Question 4.5:  (C) 35200")
    sult = SULT(udd=True).set_interest(delta=0.05)
    Z = 100000 * sult.Z_from_prob(45, prob=0.95, discrete=False)
    print(Z)

    print("SOA Question 3.9:  (E) 3850")
    sult = SULT()
    p1 = sult.p_x(20, t=25)
    p2 = sult.p_x(45, t=25)
    mean = sult.bernoulli(p1) * 2000 + sult.bernoulli(p2) * 2000
    var = (sult.bernoulli(p1, variance=True) * 2000 
           + sult.bernoulli(p2, variance=True) * 2000)
    print(sult.portfolio_percentile(mean=mean, variance=var, prob=.99))
    print()
    

    print("SOA Question 3.8:  (B) 1505")
    sult = SULT()
    p1 = sult.p_x(35, t=40)
    p2 = sult.p_x(45, t=40)
    mean = sult.bernoulli(p1) * 1000 + sult.bernoulli(p2) * 1000
    var = (sult.bernoulli(p1, variance=True) * 1000 
           + sult.bernoulli(p2, variance=True) * 1000)
    print(sult.portfolio_percentile(mean=mean, variance=var, prob=.95))
    print()
    

    print("SOA Question 3.4:  (B) 815")
    sult = SULT()
    mean = sult.p_x(25, t=95-25)
    var = sult.bernoulli(mean, variance=True)
    print(sult.portfolio_percentile(N=4000, mean=mean, variance=var, prob=.1))
    print()
    
    print("Other usage examples")
    print(sult.temporary_annuity(80, t=10)*20000) # ~130770

    E = sult.E_x(60, t=5)
    print(E, E*sult.a_x(65, benefit=(lambda x,t: 1000 + .05*t)))

    print(sult.whole_life_annuity(60))  # 14.9041
    print(sult.whole_life_annuity(60, discrete=False))  # ~13.9041
    print(sult.deferred_annuity(60, u=10))  #6.9485
    print(sult.temporary_annuity(60, t=10))  # 7.9555
    print(sult.temporary_annuity(60, t=15))  # 10.5282
    print(sult.temporary_annuity(60, t=10))
    print(sult.E_x(60, t=10))  #
    print(sult.temporary_annuity(60, t=10, discrete=False))  # ~7.5341
    print(sult.certain_life_annuity(60, u=10))  # 15.0563
    print(sult.whole_life_annuity(60, variance=True, discrete=False)) # ~10.6182
    print(sult.endowment_insurance(60, t=10)) # .62116
    print(sult.whole_life_insurance(60, moment=2)) # .10834
    print(sult.whole_life_insurance(70, moment=2)) # .21467
    print(sult.endowment_insurance(60, t=10, moment=2)) # .38732
    print(sult.temporary_annuity(60, t=10, variance=True)) #.6513


    print(sult.p_x(70)) # 0.989587
    print(math.log(sult.p_x(70, t=2)) / -2)  # 0.011103

    A1 = sult.whole_life_insurance(20, discrete=False)
    #print(A1, sult.whole_life(20))
    A2 = sult.whole_life_insurance(50, discrete=False)
    #print(A2, sult.whole_life(50))
    A3 = sult.whole_life_insurance(70, discrete=False)
    E2 = sult.E_x(20, t=30)
    E3 = sult.E_x(20, t=50)
    print(A1, E2, A2, E3, A3, 125*A1 + E2*175*A2 - E3*250*A3)  # 5,335

    print(sult.whole_life_insurance(50)) # .18931
    print(sult.term_insurance(50, t=20)) # .0402
    print(sult.endowment_insurance(50, t=20))  # ~.31471
    print(sult.E_x(50, t=20), 
          sult.whole_life_insurance(70),
          sult.deferred_insurance(50, u=20)) # .14911

    print(sult.whole_life_insurance(50, discrete=False)) # .19400
    print(sult.term_insurance(50, t=20, discrete=False)) # .0412
    print(sult.endowment_insurance(50, t=20, discrete=False))  #.38944
    print(sult.E_x(50, t=20), 
          sult.whole_life_insurance(70, discrete=False),
          sult.deferred_insurance(50, u=20, discrete=False)) # .15281
  
    print("Standard Ultimate Life Table at i=0.05")
    print(sult.frame())
    print()



"""UDD - 1/mthly insurance and annuities with uniform distribution of deaths

MIT License. Copyright 2022-2023 Terence Lim
"""
import pandas as pd
from actuarialmath.life import Interest
from actuarialmath.annuity import Annuity
from actuarialmath.mthly import Mthly

class UDD(Mthly):
    """1/mthly shortcuts with UDD assumption

    Args:
      m : number of payments per year
      life : original fractional survival and mortality functions
    """

    def __init__(self, m: int, life: Annuity, **kwargs):
        super().__init__(m=m, life=life, **kwargs)
        self.alpha_m = self.alpha(m=m, i=self.life.interest.i)
        self.beta_m = self.beta(m=m, i=self.life.interest.i)

    @staticmethod
    def alpha(m: int, i: float) -> float:
        """Derive 1/mthly UDD interest rate beta function value

        Args:
          m : number of payments per year
          i : annual interest rate
        """
        d = i / (1 + i)
        i_m = Interest.mthly(m=m, i=i)
        i_d = Interest.mthly(m=m, d=d)
        return abs(i*d / ( i_m * i_d))

    @staticmethod
    def beta(m: int, i: float) -> float:
        """Derive 1/mthly UDD interest rate alpha function value

        Args:
          m : number of payments per year
          i : annual interest rate
        """
        d = i / (1 + i)
        i_m = Interest.mthly(m=m, i=i)
        i_d = Interest.mthly(m=m, d=d)
        return abs((i - i_m) / (i_m * i_d))

    def whole_life_insurance(self, x: int, s: int = 0, moment: int = 1, 
                             b: int = 1) -> float:
        """1/mthly UDD Whole life insurance: A_x
 
        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          moment : compute first or second moment
        """
        assert moment in [1, 2, Annuity.VARIANCE]
        if moment == Annuity.VARIANCE:
            return ((self.whole_life_insurance(x, s=s, moment=2) -
                     self.whole_life_insurance(x, s=s)**2) * b**2)
        A = self.life.whole_life_insurance(x, s=s, moment=moment)
        i = self.life.interest.i
        if moment == 2:
            i = self.interest.double_force(i)
        i_m = self.life.interest.mthly(m=self.m, i=i)        
        return A * i / i_m

    def endowment_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                            endowment: int = -1, moment: int = 1) -> float:
        """1/mthly UDD Endowment insurance = term insurance + pure endowment

        Args:
          x : year of selection
          s : years after selection
          t : term of insurance in years
          b : amount of benefit
          endowment : amount of endowment
          moment : return first or second moment
        """
        assert moment in [1, 2, Annuity.VARIANCE]
        if moment == Annuity.VARIANCE:
            return (self.endowment_insurance(x, s=s, t=t, endowment=endowment, 
                                             b=b, moment=2) -
                    self.endowment_insurance(x, s=s, t=t, endowment=endowment,
                                             b=b)**2)
        E = self.E_x(x, s=s, t=t, moment=moment)
        A = self.term_insurance(x, s=s, t=t, b=b, moment=moment)
        return A  + E * (b if endowment < 0 else endowment)**moment


    def term_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                       moment: int = 1) -> float:
        """1/mthly UDD Term insurance: A_x:t

        Args:
          x : year of selection
          s : years after selection
          t : term of insurance in years
          b : amount of benefit
          moment : return first or second moment
        """
        assert moment in [1, 2, Annuity.VARIANCE]
        if moment == Annuity.VARIANCE:
            return (self.term_insurance(x, s=s, t=t, b=b, moment=2) -
                    self.term_insurance(x, s=s, t=t, b=b)**2)
        A = self.life.term_insurance(x, s=s, t=t, b=b, moment=moment)
        i = self.life.interest.i
        if moment == 2:
            i = self.interest.double_force(i)
        i_m = self.life.interest.mthly(m=self.m, i=i)        
        return A * (i / i_m)

    def whole_life_annuity(self, x: int, s: int = 0, b: int = 1, 
                           variance: bool = False) -> float:
        """1/mthly UDD Whole life annuity: a_x

        Args:
          x : year of selection
          s : years after selection
          b : amount of benefit
          variance : return first moment (False) or variance (True)
        """
        if variance:  # short cut for variance of whole life
            A1 = self.whole_life_insurance(x, s=s, moment=1)
            A2 = self.whole_life_insurance(x, s=s, moment=2)
            return (b**2 * (A2 - A1**2) / self.d**2)
        a = self.life.whole_life_annuity(x, s=s)
        return b * (a * self.alpha_m - self.beta_m)
            
    def temporary_annuity(self, x: int, s: int = 0, t: int = Annuity.WHOLE, 
                          b: int = 1, variance: bool = False) -> float:
        """1/mthly UDD Temporary life annuity: a_x:t

        Args:
          x : year of selection
          s : years after selection
          t : term of annuity in years
          b : amount of benefit
          variance : return first moment (False) or variance (True)
        """
        if variance:  # short cut for variance of temporary life annuity
            A1 = self.temporary_insurance(x, s=s, t=t)
            A2 = self.temporary_insurance(x, s=s, t=t, moment=2)
            return (b**2 * (A2 - A1**2) / self.d**2)

        # difference of whole life on (x) and deferred whole life on (x+t)
        if t < 0:
            return self.whole_life_annuity(x, b=b, s=s)  # UDD
        a = self.life.temporary_annuity(x, s=s, t=t)
        return b * (a*self.alpha_m - self.beta_m*(1 - self.E_x(x, s=s, t=t)))

    def deferred_annuity(self, x: int, s: int = 0, u: int = 0, 
                         t: int = Annuity.WHOLE, b: int = 1,
                         variance: bool = False) -> float:
        """1/mthly UDD Deferred life annuity n|t_a_x =  n+t_a_x - n_a_x

        Args:
          x : year of selection
          s : years after selection
          u : years of deferral
          t : term of annuity in years
          b : amount of benefit
        """
        if self.max_term(x+s, n) < n:
            return 0.
        if variance:  # short cut for variance of temporary life annuity
            A1 = self.endowment_insurance(x, s=s, t=t)
            A2 = self.endowment_insurance(x, s=s, t=t, moment=2)
            return (b**2 * (A2 - A1**2) / self.d**2)

        a = self.life.deferred_annuity(x, s=s, u=u, t=t)
        return (a * self.alpha_m - self.beta_m * self.E_x(x, s=s, t=u)) * b

    @staticmethod
    def interest_frame(i: float = 0.05):
        """Display 1/mthly UDD interest function values

        Args:
          i : annual interest rate
        """
        interest = Interest(i=i)
        out = pd.DataFrame(columns=["i(m)", "d(m)", "i/i(m)", "d/d(m)", 
                                    "alpha(m)", "beta(m)"],
                           index=[1, 2, 4, 12, 0], 
                           dtype=float)
        for m in out.index:
            i_m = Interest.mthly(m=m, i=interest.i)
            d_m = Interest.mthly(m=m, d=interest.d)
            out.loc[m] = [i_m, d_m, interest.i/i_m, interest.d/d_m, 
                          UDD.alpha(m=m, i=interest.i), 
                          UDD.beta(m=m, i=interest.i)]
        return out.round(5)


if __name__ == "__main__":
    from actuarialmath.sult import SULT
    from actuarialmath.policyvalues import Contract
    
    print("SOA Question 7.9:  (A) 38100")
    sult = SULT(udd=True)
    x, n, t = 45, 20, 10
    a = UDD(m=12, life=sult).temporary_annuity(x+10, t=n-10)
    print(a)
    A = UDD(m=0, life=sult).endowment_insurance(x+10, t=n-10)
    print(A)
    print(A*100000 - a*12*253)
    contract = Contract(premium=253*12, endowment=100000, benefit=100000)
    print(sult.gross_future_loss(A=A, a=a, contract=contract))
    print()

    print("SOA Question 6.49:  (C) 86")
    sult = SULT(udd=True)
    a = UDD(m=12, life=sult).temporary_annuity(40, t=20)
    A = sult.whole_life_insurance(40, discrete=False)
    P = sult.gross_premium(a=a, A=A, benefit=100000, initial_policy=200,
                           renewal_premium=0.04, initial_premium=0.04)
    print(P/12)
    print()

    from actuarialmath.recursion import Recursion    
    print("SOA Question 6.38:  (B) 11.3")
    x, n = 0, 10
    life = Recursion().set_interest(i=0.05)\
                      .set_A(0.192, x=x, t=n, endowment=1, discrete=False)\
                      .set_E(0.172, x=x, t=n)
    a = life.temporary_annuity(x, t=n, discrete=False)
    print(a)

    def fun(a):      # solve for discrete annuity, given continuous
        life = Recursion().set_interest(i=0.05)\
                          .set_a(a, x=x, t=n)\
                          .set_E(0.172, x=x, t=n)
        return UDD(m=0, life=life).temporary_annuity(x, t=n)
    a = life.solve(fun, target=a, grid=a)  # discrete annuity
    P = life.gross_premium(a=a, A=0.192, benefit=1000)
    print(P)
    print()

    print("SOA Question 6.32:  (C) 550")
    x = 0
    life = Recursion().set_interest(i=0.05).set_a(9.19, x=x)
    benefits = UDD(m=0, life=life).whole_life_insurance(x)
    payments = UDD(m=12, life=life).whole_life_annuity(x)
    print(benefits, payments)
    print(life.gross_premium(a=payments, A=benefits, benefit=100000)/12)
    print()

    print("SOA Question 6.22:  (C) 102")
    life = SULT(udd=True)
    a = UDD(m=12, life=life).temporary_annuity(45, t=20)
    A = UDD(m=0, life=life).whole_life_insurance(45)
    print(life.gross_premium(A=A, a=a, benefit=100000)/12)
    print()

    print("Interest Functions at i=0.05")
    print("----------------------------")
    print(UDD.interest_frame(i=0.05))
    print()

    



"""Base class for actuarial math, with common utility functions and constants

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
import math
import numpy as np
import scipy
from typing import Callable, Any, Tuple, List

class Actuarial(object):
    """Define constants and common utility functions

    Constants:
      VARIANCE : select variance as the statistical moment to calculate

      WHOLE : indicates that term of insurance or annuity is Whole Life
    """
    # constants
    VARIANCE = -2
    WHOLE = -999
    _VARIANCE = VARIANCE
    _WHOLE = WHOLE
    _TOL = 1e-6
    _verbose = 0
    _MAXAGE = 130    # default oldest age
    _MINAGE = 0      # default youngest age

    @classmethod
    def methods(cls, echo: bool = True):
        """Helper to partially display docstrings from classes and methods

        Args:
          echo : print if True, else return as text string
        """
        s = "\nclass " + cls.__name__ + " - "
        s += "\n".join([l for l in cls.__doc__.strip().split('\n')]) + "\n"
        if hasattr(cls, '_methods') and cls._methods:
            s += '\n    Methods:\n    --------\n\n'
            for method in cls._methods:
                meth = getattr(cls, method)
                desc = meth.__doc__.strip().split('\n')[0]
                if callable(meth):
                    args = [str(arg) for arg in signature(meth).parameters.keys()
                            if str(arg) not in ['cls', 'self']]
                else:
                    args = []
                s += f"    {method}(" + ", ".join(args) + f"):\n{' '*6}{desc}\n\n"
        return print(s) if echo else s

    #
    # Helpers for numerical computations
    #
    @staticmethod
    def integral(fun: Callable[[float], float],
                 lower: float,
                 upper: float) -> float:
        """Compute integral of the function between lower and upper limits
        
        Args:
          fun : function to integrate
          lower : lower limit
          upper : upper limit
        """
        y = scipy.integrate.quad(fun, lower, upper, full_output=1)
        return y[0]

    @staticmethod
    def derivative(fun: Callable[[float], float], x: float) -> float:
        """Compute derivative of the function at a value
        
        Args:
          fun : function to compute derivative
          x : value to compute derivative at
        """
        return scipy.misc.derivative(fun, x0=x, dx=1)

    @staticmethod
    def solve(fun: Callable[[float], float], target: float, 
              grid: float | Tuple | List, mad: bool = False) -> float:
        """Solve for the root of, or parameter value that minimizes, an equation

        Args:
          fun : function to compute output given input values
          target : target value of function output
          grid : initial guesses or bounds
          root : whether solve root, or minimize absolute difference (False)

        Returns:
          value s.t. output of function fun(value) ~ target
        """
        if mad:   # minimize absolute difference
            f = lambda t: abs(fun(t) - target)
            return scipy.optimize.minimize_scalar(f, grid).x    
        else:     # solve root
            f = lambda x: fun(x) - target
            if isinstance(grid, (list, tuple)):
                grid = min([(abs(f(x)), x)    # guess can be list of guesses
                            for x in np.linspace(min(grid), max(grid), 5)])[1]
            output = scipy.optimize.fsolve(f, [grid], full_output=True)
            fun(output[0][0])  # execute with final answer in case want side effects
            return output[0][0]

    def add_term(self, t: int, n: int) -> int:
        """Add two terms, either term may be Whole Life

        Args:
          t : first term to add
          n : second term to add
        """
        if t == self.WHOLE or n == self.WHOLE:
            return self.WHOLE  # adding any term to WHOLE is still WHOLE
        return t + n

    def max_term(self, x: int, t: int, u: int = 0) -> int:
        """Adjust term if adding term and deferral periods to (x) exceeds maxage

        Args:
          x : age
          t : term of insurance or annuity, after deferral period
          u : term deferred

        Returns:
          value of term t adjusted by deferral and maxage s.t. does not exceed maxage
        """
        if t < 0 or x + t + u > self._MAXAGE:
            return self._MAXAGE - (x + u)
        return t

if __name__ == "__main__":
    actuarial = Actuarial()
    def as_term(t): return "WHOLE_LIFE" if t == Actuarial.WHOLE else t
    
    for a,b in [(3, Actuarial.WHOLE), (3, 2), (3, -1)]:
        print(f"({as_term(a)}) + ({as_term(b)}) =",
              as_term(actuarial.add_term(a, b)))
        
    print(Actuarial.solve(fun=lambda omega: 1/omega, 
                          target=0.05,
                          grid=[1, 100]))
    print(Actuarial.derivative(fun=lambda x: x/50, x=25))
